# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

import mock
import netaddr
from neutron.common import rpc

from quark.db import api as db_api
import quark.ipam
from quark.tests.functional.base import BaseFunctionalTest


class QuarkIpamBaseFunctionalTest(BaseFunctionalTest):
    def setUp(self):
        super(QuarkIpamBaseFunctionalTest, self).setUp()

        patcher = mock.patch("neutron.common.rpc.oslo_messaging")
        patcher.start()
        self.addCleanup(patcher.stop)
        rpc.init(mock.MagicMock())


class QuarkFindSubnetAllocationCount(QuarkIpamBaseFunctionalTest):
    @contextlib.contextmanager
    def _fixtures(self, models):

        self.ipam = quark.ipam.QuarkIpamANY()
        net = dict(name="public", tenant_id='fake')
        net_mod = db_api.network_create(self.context, **net)
        with self.context.session.begin():
            for model in models:
                policy_mod = db_api.ip_policy_create(
                    self.context, **model['ip_policy'])
                model['subnet']["network"] = net_mod
                model['subnet']["ip_policy"] = policy_mod
                db_api.subnet_create(self.context, **model['subnet'])
        yield net_mod

    def _create_models(self, subnet_cidr, ip_version, next_ip):
        models = {}
        net = netaddr.IPNetwork(subnet_cidr)
        first = str(netaddr.IPAddress(net.first))
        last = str(netaddr.IPAddress(net.last))
        models['ip_policy'] = dict(name='testpolicy',
                                   description='blah',
                                   exclude=[first, last],
                                   size=2)
        models["subnet"] = dict(cidr=subnet_cidr,
                                first_ip=net.first,
                                last_ip=net.last,
                                next_auto_assign_ip=next_ip,
                                tenant_id='fake',
                                do_not_use=False)
        return models

    def _create_ip_address(self, ip_address, ip_version, subnet_cidr, net_id):
        with self.context.session.begin():
            subnet = db_api.subnet_find(self.context, None, False, None, None,
                                        cidr=subnet_cidr).all()
            ip = dict(subnet_id=subnet[0].id,
                      network_id=net_id,
                      version=ip_version,
                      address=netaddr.IPAddress(ip_address))
            db_api.ip_address_create(self.context, **ip)

    def test_ordering_subnets_find_allocation_counts_when_count_equal(self):
        models = []
        cidrs = ["0.0.0.0/31", "1.1.1.0/31", "2.2.2.0/31"]
        for cidr in cidrs:
            last = netaddr.IPNetwork(cidr).last
            models.append(self._create_models(cidr, 4, last))

        with self._fixtures(models) as net:
            subnets = db_api.subnet_find_ordered_by_most_full(
                self.context, net['id'], segment_id=None,
                scope=db_api.ALL).all()
            self.assertEqual(len(subnets), 3)
            for subnet in subnets:
                self.assertIn(subnet[0]["cidr"], cidrs)

            subnets = db_api.subnet_find_ordered_by_least_full(
                self.context, net['id'], segment_id=None,
                scope=db_api.ALL).all()
            self.assertEqual(len(subnets), 3)
            for subnet in subnets:
                self.assertIn(subnet[0]["cidr"], cidrs)

    def test_ordering_subnets_find_allocation_counts_when_counts_unequal(self):
        models = []
        cidrs = ["0.0.0.0/31", "1.1.1.0/31", "2.2.2.0/30"]
        for cidr in cidrs:
            last = netaddr.IPNetwork(cidr).last
            models.append(self._create_models(cidr, 4, last))

        with self._fixtures(models) as net:
            self._create_ip_address("2.2.2.2", 4, "2.2.2.0/30", net["id"])

            subnets_with_same_ips_used = ["0.0.0.0/31", "1.1.1.0/31"]
            subnets = db_api.subnet_find_ordered_by_most_full(
                self.context, net['id'], segment_id=None,
                scope=db_api.ALL).all()
            self.assertEqual(len(subnets), 3)
            self.assertIn(subnets[0][0].cidr, subnets_with_same_ips_used)
            self.assertEqual(subnets[0][1], 0)
            self.assertIn(subnets[1][0].cidr, subnets_with_same_ips_used)
            self.assertEqual(subnets[1][1], 0)
            self.assertEqual(subnets[2][0].cidr, "2.2.2.0/30")
            self.assertEqual(subnets[2][1], 1)

            subnets = db_api.subnet_find_ordered_by_least_full(
                self.context, net['id'], segment_id=None,
                scope=db_api.ALL).all()
            self.assertEqual(len(subnets), 3)
            self.assertEqual(subnets[0][0].cidr, "2.2.2.0/30")
            self.assertEqual(subnets[0][1], 1)
            self.assertIn(subnets[1][0].cidr, subnets_with_same_ips_used)
            self.assertEqual(subnets[1][1], 0)
            self.assertIn(subnets[2][0].cidr, subnets_with_same_ips_used)
            self.assertEqual(subnets[2][1], 0)

    def test_ordering_subnets_find_allocc_when_counts_unequal_size_equal(self):
        models = []
        cidrs = ["0.0.0.0/31", "1.1.1.0/31", "2.2.2.0/31"]
        for cidr in cidrs:
            last = netaddr.IPNetwork(cidr).last
            models.append(self._create_models(cidr, 4, last))

        with self._fixtures(models) as net:
            self._create_ip_address("2.2.2.1", 4, "2.2.2.0/31", net["id"])
            self._create_ip_address("2.2.2.2", 4, "2.2.2.0/31", net["id"])
            self._create_ip_address("1.1.1.1", 4, "1.1.1.0/31", net["id"])

            subnets = db_api.subnet_find_ordered_by_most_full(
                self.context, net['id'], segment_id=None,
                scope=db_api.ALL).all()
            self.assertEqual(len(subnets), 3)
            self.assertEqual(subnets[0][0].cidr, "2.2.2.0/31")
            self.assertEqual(subnets[0][1], 2)
            self.assertEqual(subnets[1][0].cidr, "1.1.1.0/31")
            self.assertEqual(subnets[1][1], 1)
            self.assertEqual(subnets[2][0].cidr, "0.0.0.0/31")
            self.assertEqual(subnets[2][1], 0)

            subnets = db_api.subnet_find_ordered_by_least_full(
                self.context, net['id'], segment_id=None,
                scope=db_api.ALL).all()
            self.assertEqual(len(subnets), 3)
            self.assertEqual(subnets[0][0].cidr, "0.0.0.0/31")
            self.assertEqual(subnets[0][1], 0)
            self.assertEqual(subnets[1][0].cidr, "1.1.1.0/31")
            self.assertEqual(subnets[1][1], 1)
            self.assertEqual(subnets[2][0].cidr, "2.2.2.0/31")
            self.assertEqual(subnets[2][1], 2)

    def test_ordering_subnets_ip_version(self):
        """Order by ip_version primarily.

        Order by ip_version primarily, even when IPv4 is less full than IPv6
        subnet.
        """
        cidr4 = "0.0.0.0/30"  # 2 bits
        last4 = netaddr.IPNetwork(cidr4).last
        cidr6 = "fffc::/127"  # 1 bits
        last6 = netaddr.IPNetwork(cidr6).last
        with self._fixtures([
            self._create_models(cidr4, 4, last4),
            self._create_models(cidr6, 6, last6)
        ]) as net:
            subnets = db_api.subnet_find_ordered_by_most_full(
                self.context, net['id'], segment_id=None,
                scope=db_api.ALL).all()
            self.assertEqual(subnets[0][0].ip_version, 4)
            self.assertEqual(subnets[1][0].ip_version, 6)

            subnets = db_api.subnet_find_ordered_by_least_full(
                self.context, net['id'], segment_id=None,
                scope=db_api.ALL).all()
            self.assertEqual(subnets[0][0].ip_version, 4)
            self.assertEqual(subnets[1][0].ip_version, 6)

    def test_ordering_subnets_find_unused(self):
        models = []
        cidrsv4 = ["0.0.0.0/31", "1.1.1.0/31", "2.2.2.0/31"]
        cidrsv6 = ["fffc::/127", "fffd::/127"]

        for version, cidrs in [(4, cidrsv4), (6, cidrsv6)]:
            for cidr in cidrs:
                last = netaddr.IPNetwork(cidr).last
                models.append(self._create_models(cidr, version, last))

        with self._fixtures(models) as net:
            self._create_ip_address("2.2.2.1", 4, "2.2.2.0/31", net["id"])
            self._create_ip_address("2.2.2.2", 4, "2.2.2.0/31", net["id"])
            self._create_ip_address("1.1.1.1", 4, "1.1.1.0/31", net["id"])
            self._create_ip_address("fffc::1", 6, "fffc::/127", net["id"])

            subnets = db_api.subnet_find_unused(
                self.context, net['id'], segment_id=None,
                scope=db_api.ALL).all()
            self.assertEqual(len(subnets), 2)
            self.assertEqual(subnets[0][0].cidr, "0.0.0.0/31")
            self.assertEqual(subnets[0][1], 0)
            self.assertEqual(subnets[1][0].cidr, "fffd::/127")
            self.assertEqual(subnets[1][1], 0)

    def test_subnet_set_full(self):
        cidr4 = "0.0.0.0/30"  # 2 bits
        net4 = netaddr.IPNetwork(cidr4)
        with self._fixtures([
            self._create_models(cidr4, 4, net4[0])
        ]) as net:
            subnet = db_api.subnet_find(self.context, network_id=net['id'],
                                        scope=db_api.ALL)[0]
            with self.context.session.begin():
                updated = db_api.subnet_update_set_full(self.context, subnet)
                self.context.session.refresh(subnet)
                self.assertTrue(updated)
                self.assertEqual(subnet["next_auto_assign_ip"], -1)

    def test_subnet_update_next_auto_assign_ip(self):
        cidr4 = "0.0.0.0/30"  # 2 bits
        net4 = netaddr.IPNetwork(cidr4)
        with self._fixtures([
            self._create_models(cidr4, 4, net4[0])
        ]) as net:
            subnet = db_api.subnet_find(self.context, network_id=net['id'],
                                        scope=db_api.ALL)[0]
            with self.context.session.begin():
                updated = db_api.subnet_update_next_auto_assign_ip(
                    self.context, subnet)
                self.context.session.refresh(subnet)
                self.assertTrue(updated)
                self.assertEqual(
                    netaddr.IPAddress(subnet["next_auto_assign_ip"]).ipv4(),
                    net4[1])


class QuarkFindMacAddressRangeAllocationCount(QuarkIpamBaseFunctionalTest):
    @contextlib.contextmanager
    def _fixtures(self, mac_ranges):
        self.ipam = quark.ipam.QuarkIpamANY()
        for mar in mac_ranges:
            with self.context.session.begin():
                db_api.mac_address_range_create(self.context,
                                                **mar)
        yield

    def test_mac_address_ranges(self):
        mr1_mac = netaddr.EUI("AA:AA:AA:00:00:00")
        mr1 = {"cidr": "AA:AA:AA/24", "do_not_use": False,
               "first_address": mr1_mac.value,
               "last_address": netaddr.EUI("AA:AA:AA:FF:FF:FF").value,
               "next_auto_assign_mac": mr1_mac.value}
        with self._fixtures([mr1]):
            with self.context.session.begin():
                ranges = db_api.mac_address_range_find_allocation_counts(
                    self.context)
                self.assertTrue(ranges[0]["cidr"], mr1["cidr"])

    def test_mac_address_ranges_do_not_use_returns_nothing(self):
        mr1_mac = netaddr.EUI("AA:AA:AA:00:00:00")
        mr1 = {"cidr": "AA:AA:AA/24", "do_not_use": True,
               "first_address": mr1_mac.value,
               "last_address": netaddr.EUI("AA:AA:AA:FF:FF:FF").value,
               "next_auto_assign_mac": mr1_mac.value}

        with self._fixtures([mr1]):
            with self.context.session.begin():
                ranges = db_api.mac_address_range_find_allocation_counts(
                    self.context)
                self.assertTrue(ranges is None)

    def test_mac_address_ranges_do_not_use_returns_on_use_forbidden_rage(self):
        mr1_mac = netaddr.EUI("AA:AA:AA:00:00:00")
        mr1 = {"cidr": "AA:AA:AA/24", "do_not_use": True,
               "first_address": mr1_mac.value,
               "last_address": netaddr.EUI("AA:AA:AA:FF:FF:FF").value,
               "next_auto_assign_mac": mr1_mac.value}

        with self._fixtures([mr1]):
            with self.context.session.begin():
                ranges = db_api.mac_address_range_find_allocation_counts(
                    self.context, use_forbidden_mac_range=True)
                self.assertTrue(ranges[0]["cidr"], mr1["cidr"])



"""Added transactions table

Revision ID: 5632aa202d89
Revises: 3a47813ce501
Create Date: 2015-03-18 14:54:09.061787

"""

# revision identifiers, used by Alembic.
revision = '5632aa202d89'
down_revision = '4d3ed7925db3'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('quark_transactions',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB')
    op.add_column(u'quark_ip_addresses',
                  sa.Column('transaction_id', sa.Integer(), nullable=True))
    op.create_foreign_key('fk_quark_ips_transaction_id',
                          'quark_ip_addresses',
                          'quark_transactions',
                          ['transaction_id'],
                          ['id'])


def downgrade():
    op.drop_constraint('fk_quark_ips_transaction_id', 'quark_ip_addresses',
                       type_='foreignkey')
    op.drop_column(u'quark_ip_addresses', 'transaction_id')
    op.drop_table('quark_transactions')



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import contextlib
import json

import mock
import netaddr
from neutron.api.v2 import attributes as neutron_attrs
from neutron_lib import exceptions as n_exc
from oslo_config import cfg

from quark.db import models
from quark.drivers import registry
from quark import exceptions as q_exc
from quark import network_strategy
from quark.plugin_modules import ports as quark_ports
from quark import plugin_views
from quark import tags
from quark.tests import test_quark_plugin


class TestQuarkGetPorts(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, ports=None, addrs=None):
        port_models = []
        addr_models = None
        if addrs:
            addr_models = []
            for address in addrs:
                a = models.IPAddress(**address)
                addr_models.append(a)

        if isinstance(ports, list):
            for port in ports:
                port_model = models.Port(**port)
                if addr_models:
                    port_model.ip_addresses = addr_models
                port_models.append(port_model)
        elif ports is None:
            port_models = None
        else:
            port_model = models.Port(**ports)
            if addr_models:
                port_model.ip_addresses = addr_models
            port_models = port_model

        with contextlib.nested(
            mock.patch("quark.db.api.port_find")
        ) as (port_find,):
            port_find.return_value = port_models
            yield

    def test_port_list_no_ports(self):
        with self._stubs(ports=[]):
            ports = self.plugin.get_ports(self.context, filters=None,
                                          fields=None)
            self.assertEqual(ports, [])

    def test_port_list_with_device_owner_dhcp(self):
        ip = dict(id=1, address=netaddr.IPAddress("192.168.1.100").value,
                  address_readable="192.168.1.100", subnet_id=1, network_id=2,
                  version=4)
        filters = {'network_id': ip['network_id'],
                   'device_owner': 'network:dhcp'}
        port = dict(mac_address="AA:BB:CC:DD:EE:FF", network_id=1,
                    tenant_id=self.context.tenant_id, device_id=2,
                    bridge="xenbr0", device_owner='network:dhcp')
        with self._stubs(ports=[port], addrs=[ip]):
            ports = self.plugin.get_ports(self.context, filters=filters,
                                          fields=None)
            self.assertEqual(len(ports), 1)
            self.assertEqual(ports[0]["device_owner"], "network:dhcp")

    def test_port_list_with_ports(self):
        ip = dict(id=1, address=netaddr.IPAddress("192.168.1.100").value,
                  address_readable="192.168.1.100", subnet_id=1, network_id=2,
                  version=4)
        port = dict(mac_address="AA:BB:CC:DD:EE:FF", network_id=1,
                    tenant_id=self.context.tenant_id, device_id=2,
                    bridge="xenbr0")
        expected = {'status': "ACTIVE",
                    'device_owner': None,
                    'mac_address': 'AA:BB:CC:DD:EE:FF',
                    'network_id': 1,
                    'bridge': "xenbr0",
                    'tenant_id': self.context.tenant_id,
                    'admin_state_up': None,
                    'device_id': 2}
        with self._stubs(ports=[port], addrs=[ip]):
            ports = self.plugin.get_ports(self.context, filters=None,
                                          fields=None)
            self.assertEqual(len(ports), 1)
            fixed_ips = ports[0].pop("fixed_ips")
            for key in expected.keys():
                self.assertEqual(ports[0][key], expected[key])
            self.assertEqual(fixed_ips[0]["subnet_id"], ip["subnet_id"])
            self.assertEqual(fixed_ips[0]["ip_address"],
                             ip["address_readable"])

    def test_port_show_with_int_mac(self):
        port = dict(mac_address=int('AABBCCDDEEFF', 16), network_id=1,
                    tenant_id=self.context.tenant_id, device_id=2)
        expected = {'status': "ACTIVE",
                    'device_owner': None,
                    'mac_address': 'AA:BB:CC:DD:EE:FF',
                    'network_id': 1,
                    'tenant_id': self.context.tenant_id,
                    'admin_state_up': None,
                    'fixed_ips': [],
                    'device_id': 2}
        with self._stubs(ports=port):
            result = self.plugin.get_port(self.context, 1)
            for key in expected.keys():
                self.assertEqual(result[key], expected[key])

    def test_port_show_not_found(self):
        with self._stubs(ports=None):
            with self.assertRaises(n_exc.PortNotFound):
                self.plugin.get_port(self.context, 1)

    def test_port_show_vlan_id(self):
        """Prove VLAN IDs are included in port information when available."""
        port_tags = [tags.VlanTag().serialize(5)]
        port = dict(mac_address=int('AABBCCDDEEFF', 16), network_id=1,
                    tenant_id=self.context.tenant_id, device_id=2,
                    tags=port_tags)
        expected = {'status': "ACTIVE",
                    'device_owner': None,
                    'mac_address': 'AA:BB:CC:DD:EE:FF',
                    'network_id': 1,
                    'tenant_id': self.context.tenant_id,
                    'admin_state_up': None,
                    'fixed_ips': [],
                    'device_id': 2,
                    'vlan_id': '5'}
        with self._stubs(ports=port):
            result = self.plugin.get_port(self.context, 1)
            for key in expected.keys():
                self.assertEqual(result[key], expected[key])

    def test_port_show_invalid_vlan_id(self):
        """Prove VLAN IDs are included in port information when available."""
        port_tags = [tags.VlanTag().serialize('invalid')]
        port = dict(mac_address=int('AABBCCDDEEFF', 16), network_id=1,
                    tenant_id=self.context.tenant_id, device_id=2,
                    tags=port_tags)
        expected = {'status': "ACTIVE",
                    'device_owner': None,
                    'mac_address': 'AA:BB:CC:DD:EE:FF',
                    'network_id': 1,
                    'tenant_id': self.context.tenant_id,
                    'admin_state_up': None,
                    'fixed_ips': [],
                    'device_id': 2}
        with self._stubs(ports=port):
            result = self.plugin.get_port(self.context, 1)
            for key in expected.keys():
                self.assertEqual(result[key], expected[key])


class TestQuarkGetPortsProviderSubnetIds(test_quark_plugin.TestQuarkPlugin):

    def setUp(self):
        super(TestQuarkGetPortsProviderSubnetIds, self).setUp()
        self.strategy = {
            "1": {
                "bridge": "publicnet",
                "subnets": {
                    "4": "v4-provider-subnet-id",
                    "6": "v6-provider-subnet-id"
                }
            }
        }
        self.strategy_json = json.dumps(self.strategy)
        self.old = plugin_views.STRATEGY
        plugin_views.STRATEGY = network_strategy.JSONStrategy(
            self.strategy_json)
        cfg.CONF.set_override("default_net_strategy", self.strategy_json,
                              "QUARK")

    def tearDown(self):
        plugin_views.STRATEGY = self.old

    def _port_associate_stub(self, ports, address, **kwargs):
        if not isinstance(ports, list):
            ports = [ports]

        for port in ports:
            assoc = models.PortIpAssociation()
            assoc.port_id = port.id
            assoc.ip_address_id = address.id
            assoc.port = port
            assoc.ip_address = address
            assoc.enabled = address.address_type == "fixed"
        return address

    @contextlib.contextmanager
    def _stubs(self, ports=None, addrs=None):
        port_models = []
        addr_models = None
        if addrs:
            addr_models = []
            for address in addrs:
                a = models.IPAddress(**address)
                addr_models.append(a)

        if isinstance(ports, list):
            for port in ports:
                port_model = models.Port(**port)
                if addr_models:
                    port_model.ip_addresses = addr_models
                    for addr_model in addr_models:
                        self._port_associate_stub(
                            port_model, addr_model)
                port_models.append(port_model)
        elif ports is None:
            port_models = None
        else:
            port_model = models.Port(**ports)
            if addr_models:
                port_model.ip_addresses = addr_models
                for addr_model in addr_models:
                        self._port_associate_stub(
                            port_model, addr_model)
            port_models = port_model

        with contextlib.nested(
            mock.patch("quark.db.api.port_find")
        ) as (port_find,):
            port_find.return_value = port_models
            yield

    def test_port_show_with_provider_subnet_ids(self):
        """Prove provider subnets ids are shown on the port object."""
        ip = dict(id=1, address=netaddr.IPAddress("192.168.1.100").value,
                  address_readable="192.168.1.100", subnet_id="1",
                  network_id="1", version=4, address_type="fixed")
        port = dict(mac_address=int('AABBCCDDEEFF', 16), network_id="1",
                    tenant_id=self.context.tenant_id, device_id=2)
        expected = {'status': "ACTIVE",
                    'device_owner': None,
                    'mac_address': 'AA:BB:CC:DD:EE:FF',
                    'network_id': "1",
                    'tenant_id': self.context.tenant_id,
                    'admin_state_up': None,
                    'fixed_ips': [
                        {'subnet_id': 'v4-provider-subnet-id', 'enabled': True,
                         'ip_address': '192.168.1.100'}
                    ],
                    'device_id': 2}
        with self._stubs(ports=port, addrs=[ip]):
            result = self.plugin.get_port(self.context, 1)
            for key in expected.keys():
                self.assertEqual(result[key], expected[key])

    def test_port_show_without_provider_subnet_ids(self):
        """Prove provider subnets ids are shown on the port object."""
        cfg.CONF.set_override('show_provider_subnet_ids', False, 'QUARK')
        self.addCleanup(
            cfg.CONF.clear_override, 'show_provider_subnet_ids', 'QUARK')

        ip = dict(id=1, address=netaddr.IPAddress("192.168.1.100").value,
                  address_readable="192.168.1.100", subnet_id="1",
                  network_id="1", version=4, address_type="fixed")
        port = dict(mac_address=int('AABBCCDDEEFF', 16), network_id="1",
                    tenant_id=self.context.tenant_id, device_id=2)
        expected = {'status': "ACTIVE",
                    'device_owner': None,
                    'mac_address': 'AA:BB:CC:DD:EE:FF',
                    'network_id': "1",
                    'tenant_id': self.context.tenant_id,
                    'admin_state_up': None,
                    'fixed_ips': [
                        {'subnet_id': '1', 'enabled': True,
                         'ip_address': '192.168.1.100'}
                    ],
                    'device_id': 2}
        with self._stubs(ports=port, addrs=[ip]):
            result = self.plugin.get_port(self.context, 1)
            for key in expected.keys():
                self.assertEqual(result[key], expected[key])


class TestQuarkGetPortsByIPAddress(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, ports=None, addr=None):
        addr_models = []

        for port in ports:
            ip_mod = models.IPAddress()
            ip_mod.update(addr)
            port_model = models.Port()
            port_model.update(port)

            ip_mod.ports = [port_model]
            addr_models.append(ip_mod)

        with contextlib.nested(
            mock.patch("quark.db.api.port_find_by_ip_address")
        ) as (port_find_by_addr,):
            port_find_by_addr.return_value = addr_models
            yield

    def test_port_list_by_ip_address(self):
        ip = dict(id=1, address=netaddr.IPAddress("192.168.1.100").value,
                  address_readable="192.168.1.100", subnet_id=1, network_id=2,
                  version=4)
        port = dict(mac_address="AA:BB:CC:DD:EE:FF", network_id=1,
                    tenant_id=self.context.tenant_id, device_id=2,
                    bridge="xenbr0", device_owner='network:dhcp')
        with self._stubs(ports=[port], addr=ip):
            admin_ctx = self.context.elevated()
            filters = {"ip_address": ["192.168.0.1"]}
            ports = self.plugin.get_ports(admin_ctx, filters=filters,
                                          fields=None)
            self.assertEqual(len(ports), 1)
            self.assertEqual(ports[0]["device_owner"], "network:dhcp")

    def test_port_list_by_ip_not_admin_raises(self):
        with self._stubs(ports=[]):
            filters = {"ip_address": ["192.168.0.1"]}
            with self.assertRaises(n_exc.NotAuthorized):
                self.plugin.get_ports(self.context, filters=filters,
                                      fields=None)

    def test_port_list_malformed_address_bad_request(self):
        with self._stubs(ports=[]):
            filters = {"ip_address": ["malformed-address-here"]}
            admin_ctx = self.context.elevated()
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.get_ports(admin_ctx, filters=filters, fields=None)


class TestQuarkCreatePortFailure(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, port=None, network=None, addr=None, mac=None):
        if network:
            network["network_plugin"] = "BASE"
            network["ipam_strategy"] = "ANY"
        port_model = models.Port()
        port_model.update(port)
        port_models = port_model

        with contextlib.nested(
            mock.patch("quark.db.api.port_create"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.db.api.port_find"),
            mock.patch("quark.ipam.QuarkIpam.allocate_ip_address"),
            mock.patch("quark.ipam.QuarkIpam.allocate_mac_address"),
            mock.patch("quark.db.api.port_count_all"),
        ) as (port_create, net_find, port_find, alloc_ip, alloc_mac,
              port_count):
            port_create.return_value = port_models
            net_find.return_value = network
            port_find.return_value = models.Port()
            alloc_ip.return_value = addr
            alloc_mac.return_value = mac
            port_count.return_value = 0
            yield port_create

    def test_create_multiple_ports_on_same_net_and_device_id_bad_request(self):
        network = dict(id=1, tenant_id=self.context.tenant_id)
        ip = dict()
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_1 = dict(port=dict(mac_address="AA:BB:CC:DD:EE:00", network_id=1,
                                tenant_id=self.context.tenant_id, device_id=1,
                                name="Fake"))
        port_2 = dict(port=dict(mac_address="AA:BB:CC:DD:EE:11", network_id=1,
                                tenant_id=self.context.tenant_id, device_id=1,
                                name="Faker"))

        with self._stubs(port=port_1, network=network, addr=ip, mac=mac):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_port(self.context, port_1)
                self.plugin.create_port(self.context, port_2)


class TestQuarkCreatePortRM9305(test_quark_plugin.TestQuarkPlugin):
    def setUp(self):
        super(TestQuarkCreatePortRM9305, self).setUp()
        strategy = {"00000000-0000-0000-0000-000000000000":
                    {"bridge": "publicnet",
                     "subnets": {"4": "public_v4",
                                 "6": "public_v6"}},
                    "11111111-1111-1111-1111-111111111111":
                    {"bridge": "servicenet",
                     "subnets": {"4": "private_v4",
                                 "6": "private_v6"}}}
        strategy_json = json.dumps(strategy)
        quark_ports.STRATEGY = network_strategy.JSONStrategy(strategy_json)

    @contextlib.contextmanager
    def _stubs(self, port=None, network=None, addr=None, mac=None):
        if network:
            network["network_plugin"] = "BASE"
            network["ipam_strategy"] = "ANY"
        port_model = models.Port()
        port_model.update(port)
        port_models = port_model
        db_mod = "quark.db.api"
        ipam = "quark.ipam.QuarkIpam"
        with contextlib.nested(
            mock.patch("%s.port_create" % db_mod),
            mock.patch("%s.network_find" % db_mod),
            mock.patch("%s.port_find" % db_mod),
            mock.patch("%s.allocate_ip_address" % ipam),
            mock.patch("%s.allocate_mac_address" % ipam),
            mock.patch("%s.port_count_all" % db_mod),
        ) as (port_create, net_find, port_find, alloc_ip, alloc_mac,
              port_count):
            port_create.return_value = port_models
            net_find.return_value = network
            port_find.return_value = None
            alloc_ip.return_value = addr
            alloc_mac.return_value = mac
            port_count.return_value = 0
            yield port_create

    def test_RM9305_tenant_create_servicenet_port(self):
        network_id = "11111111-1111-1111-1111-111111111111"
        network = dict(id=network_id,
                       tenant_id="rackspace")
        ip = dict()
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_1 = dict(port=dict(mac_address="AA:BB:CC:DD:EE:00",
                                network_id=network_id,
                                tenant_id=self.context.tenant_id, device_id=2,
                                segment_id="bar",
                                name="Fake"))

        with self._stubs(port=port_1, network=network, addr=ip, mac=mac):
            self.plugin.create_port(self.context, port_1)

    def test_RM9305_tenant_create_publicnet_port(self):
        network_id = "00000000-0000-0000-0000-000000000000"
        network = dict(id=network_id,
                       tenant_id="rackspace")
        ip = dict()
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_1 = dict(port=dict(mac_address="AA:BB:CC:DD:EE:00",
                                network_id=network_id,
                                tenant_id=self.context.tenant_id, device_id=3,
                                segment_id="bar",
                                name="Fake"))

        with self._stubs(port=port_1, network=network, addr=ip, mac=mac):
            self.plugin.create_port(self.context, port_1)

    def test_RM9305_tenant_create_tenants_port(self):
        network_id = "foobar"
        network = dict(id=network_id,
                       tenant_id=self.context.tenant_id)
        ip = dict()
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_1 = dict(port=dict(mac_address="AA:BB:CC:DD:EE:00",
                                network_id=network_id,
                                tenant_id=self.context.tenant_id, device_id=4,
                                name="Fake"))

        with self._stubs(port=port_1, network=network, addr=ip, mac=mac):
            self.plugin.create_port(self.context, port_1)

    def test_RM9305_tenant_create_other_tenants_port(self):
        network_id = "foobar"
        network = dict(id=network_id,
                       tenant_id="other_tenant")
        ip = dict()
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_1 = dict(port=dict(mac_address="AA:BB:CC:DD:EE:00",
                                network_id=network_id,
                                tenant_id=self.context.tenant_id, device_id=5,
                                name="Fake"))

        with self._stubs(port=port_1, network=network, addr=ip, mac=mac):
            with self.assertRaises(n_exc.NotAuthorized):
                self.plugin.create_port(self.context, port_1)


class TestQuarkCreatePortsSameDevBadRequest(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, port=None, network=None, addr=None, mac=None,
               limit_checks=None, subnet=None):
        subnet_model = None

        if subnet:
            subnet_model = models.Subnet()
            subnet_model.update(subnet)

        if network:
            network["network_plugin"] = "BASE"
            network["ipam_strategy"] = "ANY"

        def _create_db_port(context, **kwargs):
            port_model = models.Port()
            port_model.update(kwargs)
            return port_model

        def _alloc_ip(context, new_ips, *args, **kwargs):
            ip_mod = models.IPAddress()
            ip_mod.update(addr)
            ip_mod.enabled_for_port = lambda x: True
            new_ips.extend([ip_mod])
            return mock.DEFAULT

        with contextlib.nested(
            mock.patch("quark.db.api.port_create"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.ipam.QuarkIpam.allocate_ip_address"),
            mock.patch("quark.ipam.QuarkIpam.allocate_mac_address"),
            mock.patch("quark.db.api.port_count_all"),
            mock.patch("neutron.quota.QuotaEngine.limit_check"),
            mock.patch("quark.db.api.subnet_find"),
        ) as (port_create, net_find, alloc_ip, alloc_mac, port_count,
              limit_check, subnet_find):
            port_create.side_effect = _create_db_port
            net_find.return_value = network
            alloc_ip.side_effect = _alloc_ip
            alloc_mac.return_value = mac
            if subnet:
                subnet_find.return_value = [subnet_model]
            port_count.return_value = 0
            if limit_checks:
                limit_check.side_effect = limit_checks
            yield port_create

    def test_create_port(self):
        network = dict(id=1, tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name))

        expected = {'status': "ACTIVE",
                    'name': port_name,
                    'device_owner': None,
                    'mac_address': mac["address"],
                    'network_id': network["id"],
                    'tenant_id': self.context.tenant_id,
                    'admin_state_up': None,
                    'fixed_ips': [],
                    'device_id': 2}
        with self._stubs(port=port["port"], network=network, addr=ip,
                         mac=mac) as port_create:
            result = self.plugin.create_port(self.context, port)
            self.assertTrue(port_create.called)
            for key in expected.keys():
                self.assertEqual(result[key], expected[key])

    def test_create_port_segment_id_on_unshared_net_ignored(self):
        network = dict(id=1, tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              segment_id="cell01", name=port_name))
        expected = {'status': "ACTIVE",
                    'name': port_name,
                    'device_owner': None,
                    'mac_address': mac["address"],
                    'network_id': network["id"],
                    'tenant_id': self.context.tenant_id,
                    'admin_state_up': None,
                    'fixed_ips': [],
                    'device_id': 2}
        with self._stubs(port=port["port"], network=network, addr=ip,
                         mac=mac) as port_create:
            result = self.plugin.create_port(self.context, port)
            self.assertTrue(port_create.called)
            for key in expected.keys():
                self.assertEqual(result[key], expected[key])

    def test_create_port_mac_address_not_specified(self):
        network = dict(id=1, tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2))
        expected = {'status': "ACTIVE",
                    'device_owner': None,
                    'mac_address': mac["address"],
                    'network_id': network["id"],
                    'tenant_id': self.context.tenant_id,
                    'admin_state_up': None,
                    'fixed_ips': [],
                    'device_id': 2}
        with self._stubs(port=port["port"], network=network, addr=ip,
                         mac=mac) as port_create:
            port["port"]["mac_address"] = neutron_attrs.ATTR_NOT_SPECIFIED
            result = self.plugin.create_port(self.context, port)
            self.assertTrue(port_create.called)
            for key in expected.keys():
                self.assertEqual(result[key], expected[key])

    @mock.patch("quark.network_strategy.JSONStrategy.is_provider_network")
    def test_create_providernet_port_fixed_ip_not_authorized(self, is_parent):
        is_parent.return_value = True
        network = dict(id='1', tenant_id=self.context.tenant_id)
        subnet = dict(id=1, network_id=network["id"])
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        ip = mock.MagicMock()
        ip.get = lambda x, *y: 1 if x == "subnet_id" else None
        ip.formatted = lambda: "192.168.10.45"
        ip.enabled_for_port = lambda x: True
        fixed_ips = [dict(subnet_id=1, enabled=True,
                     ip_address="192.168.10.45")]
        port = dict(port=dict(mac_address=mac["address"], network_id='1',
                              tenant_id=self.context.tenant_id, device_id=2,
                              fixed_ips=fixed_ips, ip_addresses=[ip],
                              segment_id="provider_segment"))

        with self._stubs(port=port["port"], network=network, addr=ip,
                         mac=mac, subnet=subnet):
            with self.assertRaises(n_exc.NotAuthorized):
                self.plugin.create_port(self.context, port)

    @mock.patch("quark.network_strategy.JSONStrategy.is_provider_network")
    def test_create_providernet_port_fixed_ip_wrong_segment(self, is_parent):
        is_parent.return_value = True
        network = dict(id='1', tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        subnet = dict(id=1, network_id=network["id"])
        ip = mock.MagicMock()
        ip.get = lambda x, *y: 1 if x == "subnet_id" else None
        ip.formatted = lambda: "192.168.10.45"
        ip.enabled_for_port = lambda x: True
        fixed_ips = [dict(subnet_id=1, enabled=True,
                     ip_address="192.168.10.45")]
        port = dict(port=dict(mac_address=mac["address"], network_id='1',
                              tenant_id=self.context.tenant_id, device_id=2,
                              fixed_ips=fixed_ips, ip_addresses=[ip],
                              segment_id="provider_segment"))

        with self._stubs(port=port["port"], network=network, addr=ip,
                         mac=mac, subnet=subnet):
            with self.assertRaises(q_exc.AmbiguousNetworkId):
                self.plugin.create_port(self.context.elevated(), port)

    def test_create_port_fixed_ip_subnet_not_found(self):
        network = dict(id='1', tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        ip = mock.MagicMock()
        ip.get = lambda x, *y: 1 if x == "subnet_id" else None
        ip.formatted = lambda: "192.168.10.45"
        ip.enabled_for_port = lambda x: True
        fixed_ips = [dict(subnet_id=1, enabled=True,
                     ip_address="192.168.10.45")]
        port = dict(port=dict(mac_address=mac["address"], network_id='1',
                              tenant_id=self.context.tenant_id, device_id=2,
                              fixed_ips=fixed_ips, ip_addresses=[ip],
                              segment_id="provider_segment"))

        with self._stubs(port=port["port"], network=network, addr=ip,
                         mac=mac):
            with self.assertRaises(n_exc.NotFound):
                self.plugin.create_port(self.context.elevated(), port)

    def test_create_port_fixed_ip_subnet_not_in_network(self):
        network = dict(id='1', tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        subnet = dict(id=1, network_id='2')

        ip = mock.MagicMock()
        ip.get = lambda x, *y: 1 if x == "subnet_id" else None
        ip.formatted = lambda: "192.168.10.45"
        ip.enabled_for_port = lambda x: True
        fixed_ips = [dict(subnet_id=1, enabled=True,
                     ip_address="192.168.10.45")]
        port = dict(port=dict(mac_address=mac["address"], network_id='1',
                              tenant_id=self.context.tenant_id, device_id=2,
                              fixed_ips=fixed_ips, ip_addresses=[ip],
                              segment_id="provider_segment"))

        with self._stubs(port=port["port"], network=network, addr=ip,
                         mac=mac, subnet=subnet):
            with self.assertRaises(n_exc.InvalidInput):
                self.plugin.create_port(self.context.elevated(), port)

    def test_create_port_fixed_ips_bad_request(self):
        network = dict(id=1, tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        ip = mock.MagicMock()
        ip.get = lambda x: 1 if x == "subnet_id" else None
        ip.formatted = lambda: "192.168.10.45"
        fixed_ips = [dict()]
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              fixed_ips=fixed_ips, ip_addresses=[ip]))
        with self._stubs(port=port["port"], network=network, addr=ip,
                         mac=mac):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_port(self.context, port)

    def test_create_port_no_network_found(self):
        port = dict(port=dict(network_id=1, tenant_id=self.context.tenant_id,
                              device_id=2))
        with self._stubs(network=None, port=port["port"]):
            with self.assertRaises(n_exc.NetworkNotFound):
                self.plugin.create_port(self.context, port)

    def test_create_port_security_groups_raises(self, groups=[1]):
        network = dict(id=1, tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()
        group = models.SecurityGroup()
        group.update({'id': 1, 'tenant_id': self.context.tenant_id,
                      'name': 'foo', 'description': 'bar'})
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, security_groups=[group]))
        with self._stubs(port=port["port"], network=network, addr=ip,
                         mac=mac):
            with mock.patch("quark.db.api.security_group_find"):
                with self.assertRaises(q_exc.SecurityGroupsNotImplemented):
                    self.plugin.create_port(self.context, port)


class TestQuarkPortCreateQuota(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, port=None, network=None, addr=None, mac=None):
        if network:
            network["network_plugin"] = "BASE"
            network["ipam_strategy"] = "ANY"
        port_model = models.Port()
        port_model.update(port)
        port_models = port_model

        with contextlib.nested(
            mock.patch("quark.db.api.port_create"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.ipam.QuarkIpam.allocate_ip_address"),
            mock.patch("quark.ipam.QuarkIpam.allocate_mac_address"),
            mock.patch("quark.db.api.port_count_all"),
            mock.patch("neutron.quota.QuotaEngine.limit_check")
        ) as (port_create, net_find, alloc_ip, alloc_mac, port_count,
              limit_check):
            port_create.return_value = port_models
            net_find.return_value = network
            alloc_ip.return_value = addr
            alloc_mac.return_value = mac
            port_count.return_value = len(network["ports"])
            limit_check.side_effect = n_exc.OverQuota
            yield port_create

    def test_create_port_net_at_max(self):
        network = dict(id=1, ports=[models.Port()],
                       tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name))
        with self._stubs(port=port["port"], network=network, addr=ip, mac=mac):
            with self.assertRaises(n_exc.OverQuota):
                self.plugin.create_port(self.context, port)


class TestQuarkPortCreateFixedIpsQuota(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, network):
        network["network_plugin"] = "BASE"
        network["ipam_strategy"] = "ANY"
        with mock.patch("quark.db.api.network_find") as net_find:
            net_find.return_value = network
            yield

    def test_create_port_fixed_ips_over_quota(self):
        network = {"id": 1, "tenant_id": self.context.tenant_id}
        fixed_ips = [{"subnet_id": 1}, {"subnet_id": 1}, {"subnet_id": 1},
                     {"subnet_id": 1}, {"subnet_id": 1}, {"subnet_id": 1},
                     {"subnet_id": 1}]
        port = {"port": {"network_id": 1, "tenant_id": self.context.tenant_id,
                         "device_id": 2, "fixed_ips": fixed_ips}}
        with self._stubs(network=network):
            with self.assertRaises(n_exc.OverQuota):
                self.plugin.create_port(self.context, port)


class TestQuarkUpdatePort(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, port, new_ips=None, parent_net=False):
        port_model = None
        if port:
            net_model = models.Network()
            net_model["network_plugin"] = "BASE"
            port_model = models.Port()
            port_model.network = net_model
            port_model.update(port)

        with contextlib.nested(
            mock.patch("quark.db.api.port_find"),
            mock.patch("quark.db.api.port_update"),
            mock.patch("quark.ipam.QuarkIpam.allocate_ip_address"),
            mock.patch("quark.ipam.QuarkIpam.deallocate_ips_by_port"),
        ) as (port_find, port_update, alloc_ip, dealloc_ip):
            port_find.return_value = port_model
            port_update.return_value = port_model
            if new_ips:
                alloc_ip.return_value = new_ips
            yield port_find, port_update, alloc_ip, dealloc_ip

    def test_update_port_not_found(self):
        with self._stubs(port=None):
            with self.assertRaises(n_exc.PortNotFound):
                self.plugin.update_port(self.context, 1, {})

    def test_update_port(self):
        with self._stubs(
            port=dict(id=1, name="myport")
        ) as (port_find, port_update, alloc_ip, dealloc_ip):
            new_port = dict(port=dict(name="ourport"))
            self.plugin.update_port(self.context, 1, new_port)
            self.assertEqual(port_find.call_count, 2)
            port_update.assert_called_once_with(
                self.context,
                port_find(),
                name="ourport",
                security_groups=[])

    def test_update_port_fixed_ip_bad_request(self):
        with self._stubs(
            port=dict(id=1, name="myport")
        ) as (port_find, port_update, alloc_ip, dealloc_ip):
            new_port = dict(port=dict(
                fixed_ips=[dict(subnet_id=None,
                                ip_address=None)]))
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.update_port(self.context, 1, new_port)

    def test_update_port_fixed_ip_bad_request_malformed_address(self):
        with self._stubs(
            port=dict(id=1, name="myport", mac_address="0:0:0:0:0:1")
        ) as (port_find, port_update, alloc_ip, dealloc_ip):
            new_port = dict(port=dict(
                fixed_ips=[dict(subnet_id=1,
                                ip_address="malformed-address-here")]))
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.update_port(self.context, 1, new_port)

    def test_update_port_fixed_ip(self):
        with self._stubs(
            port=dict(id=1, name="myport", mac_address="0:0:0:0:0:1")
        ) as (port_find, port_update, alloc_ip, dealloc_ip):
            new_port = dict(port=dict(
                fixed_ips=[dict(subnet_id=1,
                                ip_address="1.1.1.1")]))
            self.plugin.update_port(self.context, 1, new_port)
            self.assertEqual(alloc_ip.call_count, 1)

    def test_update_port_fixed_ip_no_subnet_raises(self):
        with self._stubs(
            port=dict(id=1, name="myport", mac_address="0:0:0:0:0:1")
        ) as (port_find, port_update, alloc_ip, dealloc_ip):
            new_port = dict(port=dict(
                fixed_ips=[dict(ip_address="1.1.1.1")]))
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.update_port(self.context, 1, new_port)

    def test_update_port_fixed_ip_subnet_only_allocates_ip(self):
        with self._stubs(
            port=dict(id=1, name="myport", mac_address="0:0:0:0:0:1")
        ) as (port_find, port_update, alloc_ip, dealloc_ip):
            new_port = dict(port=dict(
                fixed_ips=[dict(subnet_id=1)]))
            self.plugin.update_port(self.context, 1, new_port)
            self.assertEqual(alloc_ip.call_count, 1)

    def test_update_port_fixed_ip_allocs_new_deallocs_existing(self):
        addr_dict = {"address": 0, "address_readable": "0.0.0.0"}
        addr = models.IPAddress()
        addr.update(addr_dict)
        new_addr_dict = {"address": netaddr.IPAddress("1.1.1.1"),
                         "address_readable": "1.1.1.1"}
        new_addr = models.IPAddress()
        new_addr.update(new_addr_dict)

        with self._stubs(
            port=dict(id=1, name="myport", mac_address="0:0:0:0:0:1",
                      ip_addresses=[addr]),
            new_ips=[new_addr]
        ) as (port_find, port_update, alloc_ip, dealloc_ip):
            new_port = dict(port=dict(
                fixed_ips=[dict(subnet_id=1,
                                ip_address=new_addr["address_readable"])]))
            self.plugin.update_port(self.context, 1, new_port)
            self.assertEqual(alloc_ip.call_count, 1)

    def test_update_port_goes_over_quota(self):
        fixed_ips = {"fixed_ips": [{"subnet_id": 1},
                                   {"subnet_id": 1},
                                   {"subnet_id": 1},
                                   {"subnet_id": 1},
                                   {"subnet_id": 1},
                                   {"subnet_id": 1},
                                   {"subnet_id": 1}]}
        with self._stubs(
            port=dict(id=1, name="myport", mac_address="0:0:0:0:0:1")
        ) as (port_find, port_update, alloc_ip, dealloc_ip):
            new_port = {"port": fixed_ips}
            with self.assertRaises(n_exc.OverQuota):
                self.plugin.update_port(self.context, 1, new_port)


class TestQuarkUpdatePortSecurityGroups(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, port, new_ips=None, parent_net=False):
        port_model = None
        sg_mod = models.SecurityGroup()
        if port:
            net_model = models.Network()
            net_model["network_plugin"] = "BASE"
            port_model = models.Port()
            port_model.network = net_model
            port_model.update(port)
            port_model["security_groups"].append(sg_mod)

        with contextlib.nested(
            mock.patch("quark.db.api.port_find"),
            mock.patch("quark.db.api.port_update"),
            mock.patch("quark.ipam.QuarkIpam.allocate_ip_address"),
            mock.patch("quark.ipam.QuarkIpam.deallocate_ips_by_port"),
            mock.patch("neutron.quota.QuotaEngine.limit_check"),
            mock.patch("quark.plugin_modules.ports.STRATEGY"
                       ".is_provider_network"),
            mock.patch("quark.db.api.security_group_find"),
            mock.patch("quark.drivers.base.BaseDriver.update_port")
        ) as (port_find, port_update, alloc_ip, dealloc_ip, limit_check,
              net_strat, sg_find, driver_port_update):
            port_find.return_value = port_model

            def _port_update(context, port_db, **kwargs):
                return port_db.update(kwargs)
            port_update.side_effect = _port_update
            if new_ips:
                alloc_ip.return_value = new_ips
            net_strat.return_value = parent_net
            sg_find.return_value = sg_mod
            yield (port_find, port_update, alloc_ip, dealloc_ip, sg_find,
                   driver_port_update)

    def test_update_port_security_groups(self):
        with self._stubs(
            port=dict(id=1, device_id="device"), parent_net=True
        ) as (port_find, port_update, alloc_ip, dealloc_ip, sg_find,
              driver_port_update):
            new_port = dict(port=dict(name="ourport",
                                      security_groups=[1]))
            port = self.plugin.update_port(self.context, 1, new_port)
            port_update.assert_called_once_with(
                self.context,
                port_find(),
                name="ourport",
                security_groups=[sg_find()])
            self.assertEqual(sg_find()["id"], port["security_groups"][0])

    def test_update_port_empty_list_security_groups(self):
        port_dict = {"id": 1, "mac_address": "AA:BB:CC:DD:EE:FF",
                     "device_id": 2, "backend_key": 3}
        with self._stubs(
            port=port_dict, parent_net=True
        ) as (port_find, port_update, alloc_ip, dealloc_ip, sg_find,
              driver_port_update):
            new_port = dict(port=dict(name="ourport",
                                      security_groups=[]))
            port = self.plugin.update_port(self.context, 1, new_port)
            self.assertEqual(port["security_groups"], [])
            port_update.assert_called_once_with(
                self.context,
                port_find(),
                name="ourport",
                security_groups=[])
            driver_port_update.assert_called_once_with(
                self.context, port_id=port_dict["backend_key"],
                mac_address=port_dict["mac_address"],
                device_id=port_dict["device_id"],
                security_groups=[],
                base_net_driver=registry.DRIVER_REGISTRY.get_driver('BASE'))

    def test_update_port_no_security_groups(self):
        port_dict = {"id": 1, "mac_address": "AA:BB:CC:DD:EE:FF",
                     "device_id": 2, "backend_key": 3}
        with self._stubs(
            port=port_dict, parent_net=True
        ) as (port_find, port_update, alloc_ip, dealloc_ip, sg_find,
              driver_port_update):
            new_port = dict(port=dict(name="ourport"))
            self.plugin.update_port(self.context, 1, new_port)
            driver_port_update.assert_called_once_with(
                self.context, port_id=port_dict["backend_key"],
                mac_address=port_dict["mac_address"],
                device_id=port_dict["device_id"],
                base_net_driver=registry.DRIVER_REGISTRY.get_driver('BASE'))

    def test_update_port_security_groups_no_device_id_raises(self):
        with self._stubs(
            port=dict(id=1), parent_net=True
        ) as (port_find, port_update, alloc_ip, dealloc_ip, sg_find,
              driver_port_update):
            new_port = dict(port=dict(name="ourport",
                                      security_groups=[1]))
            with self.assertRaises(q_exc.SecurityGroupsRequireDevice):
                self.plugin.update_port(self.context, 1, new_port)


class TestQuarkUpdatePortSetsIps(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, port, new_ips=None):
        def alloc_mock(kls, context, addresses, *args, **kwargs):
            addresses.extend(new_ips)
            self.called = True

        port_model = None
        if port:
            net_model = models.Network()
            net_model["network_plugin"] = "BASE"
            port_model = models.Port()
            port_model['network'] = net_model
            port_model.update(port)
        with contextlib.nested(
            mock.patch("quark.db.api.port_find"),
            mock.patch("quark.db.api.port_update"),
            mock.patch("quark.ipam.QuarkIpam.deallocate_ips_by_port"),
            mock.patch("neutron.quota.QuotaEngine.limit_check")
        ) as (port_find, port_update, dealloc_ip, limit_check):
            port_find.return_value = port_model
            port_update.return_value = port_model
            alloc_ip = mock.patch("quark.ipam.QuarkIpam.allocate_ip_address",
                                  new=alloc_mock)
            alloc_ip.start()
            yield port_find, port_update, alloc_ip, dealloc_ip
            alloc_ip.stop()

    def test_update_port_fixed_ip_subnet_only_allocates_ip(self):
        self.called = False
        new_addr_dict = {"address": netaddr.IPAddress('1.1.1.1'),
                         "address_readable": "1.1.1.1"}
        new_addr = models.IPAddress()
        new_addr.update(new_addr_dict)
        with self._stubs(
            port=dict(id=1, name="myport", mac_address="0:0:0:0:0:1"),
            new_ips=[new_addr]
        ) as (port_find, port_update, alloc_ip, dealloc_ip):
            new_port = dict(port=dict(
                fixed_ips=[dict(subnet_id=1)]))
            self.plugin.update_port(self.context, 1, new_port)
            self.assertTrue(self.called)


class TestQuarkCreatePortOnSharedNetworks(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, port=None, network=None, addr=None, mac=None):
        self.strategy = {"public_network":
                         {"bridge": "xenbr0",
                          "subnets": {"4": "public_v4",
                                      "6": "public_v6"}}}
        strategy_json = json.dumps(self.strategy)
        quark_ports.STRATEGY = network_strategy.JSONStrategy(strategy_json)

        if network:
            network["network_plugin"] = "BASE"
            network["ipam_strategy"] = "ANY"
        port_model = models.Port()
        port_model.update(port)
        port_models = port_model

        with contextlib.nested(
            mock.patch("quark.db.api.port_create"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.ipam.QuarkIpam.allocate_ip_address"),
            mock.patch("quark.ipam.QuarkIpam.allocate_mac_address"),
            mock.patch("neutron.quota.QuotaEngine.limit_check")
        ) as (port_create, net_find, alloc_ip, alloc_mac, limit_check):
            port_create.return_value = port_models
            net_find.return_value = network
            alloc_ip.return_value = addr
            alloc_mac.return_value = mac
            yield port_create

    def test_create_port_shared_net_no_quota_check(self):
        network = dict(id=1, ports=[models.Port()],
                       tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"],
                              network_id="public_network",
                              tenant_id=self.context.tenant_id, device_id=2,
                              segment_id="cell01",
                              name=port_name))
        with self._stubs(port=port["port"], network=network, addr=ip, mac=mac):
            try:
                self.plugin.create_port(self.context, port)
            except Exception:
                self.fail("create_port raised OverQuota")

    def test_create_port_shared_net_no_segment_id_fails(self):
        network = dict(id=1, ports=[models.Port()],
                       tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"],
                              network_id="public_network",
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name))
        with self._stubs(port=port["port"], network=network, addr=ip, mac=mac):
            with self.assertRaises(q_exc.AmbiguousNetworkId):
                self.plugin.create_port(self.context, port)


class TestQuarkGetPortCount(test_quark_plugin.TestQuarkPlugin):
    def test_get_port_count(self):
        """This isn't really testable."""
        with mock.patch("quark.db.api.port_count_all"):
            self.plugin.get_ports_count(self.context, {})


class TestQuarkDeletePort(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, port=None, addr=None, mac=None):
        port_models = None
        if port:
            net_model = models.Network()
            net_model["network_plugin"] = "BASE"
            net_model["ipam_strategy"] = "ANY"
            port_model = models.Port()
            port_model.update(port)
            port_model.network = net_model
            port_models = port_model

        with contextlib.nested(
            mock.patch("quark.db.api.port_find"),
            mock.patch("quark.ipam.QuarkIpam.deallocate_ips_by_port"),
            mock.patch("quark.ipam.QuarkIpam.deallocate_mac_address"),
            mock.patch("quark.db.api.port_delete"),
            mock.patch("quark.drivers.base.BaseDriver.delete_port")
        ) as (port_find, dealloc_ip, dealloc_mac, db_port_del,
              driver_port_del):
            port_find.return_value = port_models
            dealloc_ip.return_value = addr
            dealloc_mac.return_value = mac
            yield db_port_del, driver_port_del

    def test_port_delete(self):
        port = dict(port=dict(network_id=1, tenant_id=self.context.tenant_id,
                              device_id=2, mac_address="AA:BB:CC:DD:EE:FF",
                              backend_key="foo"))
        with self._stubs(port=port["port"]) as (db_port_del, driver_port_del):
            self.plugin.delete_port(self.context, 1)
            self.assertTrue(db_port_del.called)
            driver_port_del.assert_called_with(
                self.context, "foo", mac_address=port["port"]["mac_address"],
                device_id=port["port"]["device_id"],
                base_net_driver=registry.DRIVER_REGISTRY.get_driver("BASE"))

    def test_port_delete_port_not_found_fails(self):
        with self._stubs(port=None) as (db_port_del, driver_port_del):
            with self.assertRaises(n_exc.PortNotFound):
                self.plugin.delete_port(self.context, 1)


class TestPortDiagnose(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, port, list_format=False):
        port_res = None
        if port:
            network_mod = models.Network()
            port_mod = models.Port()
            port_mod.update(port)
            network_mod["network_plugin"] = "UNMANAGED"
            port_mod.network = network_mod
            port_res = port_mod
            if list_format:
                ports = mock.MagicMock()
                ports.all.return_value = [port_mod]
                port_res = ports

        with mock.patch("quark.db.api.port_find") as port_find:
            port_find.return_value = port_res
            yield

    def test_port_diagnose(self):
        ip = dict(id=1, address=netaddr.IPAddress("192.168.1.100"),
                  address_readable="192.168.1.100", subnet_id=1, network_id=2,
                  version=4)
        fixed_ips = [{"subnet_id": ip["subnet_id"],
                      "ip_address": ip["address_readable"]}]
        port = dict(port=dict(network_id=1, tenant_id=self.context.tenant_id,
                              device_id=2, mac_address="AA:BB:CC:DD:EE:FF",
                              backend_key="foo", fixed_ips=fixed_ips,
                              network_plugin="UNMANAGED"))
        with self._stubs(port=port):
            diag = self.plugin.diagnose_port(self.context.elevated(), 1, [])
            ports = diag["ports"]
            # All none because we're using the unmanaged driver, which
            # doesn't do anything with these
            self.assertEqual(ports["status"], "ACTIVE")
            self.assertEqual(ports["device_owner"], None)
            self.assertEqual(ports["fixed_ips"], [])
            self.assertEqual(ports["security_groups"], [])
            self.assertEqual(ports["device_id"], None)
            self.assertEqual(ports["admin_state_up"], None)
            self.assertEqual(ports["network_id"], None)
            self.assertEqual(ports["tenant_id"], None)
            self.assertEqual(ports["mac_address"], None)

    def test_port_diagnose_with_wildcard(self):
        ip = dict(id=1, address=netaddr.IPAddress("192.168.1.100"),
                  address_readable="192.168.1.100", subnet_id=1, network_id=2,
                  version=4)
        fixed_ips = [{"subnet_id": ip["subnet_id"],
                      "ip_address": ip["address_readable"]}]
        port = dict(port=dict(network_id=1, tenant_id=self.context.tenant_id,
                              device_id=2, mac_address="AA:BB:CC:DD:EE:FF",
                              backend_key="foo", fixed_ips=fixed_ips,
                              network_plugin="UNMANAGED"))
        with self._stubs(port=port, list_format=True):
            diag = self.plugin.diagnose_port(self.context.elevated(), '*', [])
            ports = diag["ports"]
            # All none because we're using the unmanaged driver, which
            # doesn't do anything with these
            self.assertEqual(ports[0]["status"], "ACTIVE")
            self.assertEqual(ports[0]["device_owner"], None)
            self.assertEqual(ports[0]["fixed_ips"], [])
            self.assertEqual(ports[0]["security_groups"], [])
            self.assertEqual(ports[0]["device_id"], None)
            self.assertEqual(ports[0]["admin_state_up"], None)
            self.assertEqual(ports[0]["network_id"], None)
            self.assertEqual(ports[0]["tenant_id"], None)
            self.assertEqual(ports[0]["mac_address"], None)

    def test_port_diagnose_with_config_field(self):
        ip = dict(id=1, address=netaddr.IPAddress("192.168.1.100"),
                  address_readable="192.168.1.100", subnet_id=1, network_id=2,
                  version=4)
        fixed_ips = [{"subnet_id": ip["subnet_id"],
                      "ip_address": ip["address_readable"]}]
        port = dict(port=dict(network_id=1, tenant_id=self.context.tenant_id,
                              device_id=2, mac_address="AA:BB:CC:DD:EE:FF",
                              backend_key="foo", fixed_ips=fixed_ips,
                              network_plugin="UNMANAGED"))
        with self._stubs(port=port, list_format=True):
            diag = self.plugin.diagnose_port(self.context.elevated(), '*',
                                             ["config"])
            ports = diag["ports"]
            # All none because we're using the unmanaged driver, which
            # doesn't do anything with these
            self.assertEqual(ports[0]["status"], "ACTIVE")
            self.assertEqual(ports[0]["device_owner"], None)
            self.assertEqual(ports[0]["fixed_ips"], [])
            self.assertEqual(ports[0]["security_groups"], [])
            self.assertEqual(ports[0]["device_id"], None)
            self.assertEqual(ports[0]["admin_state_up"], None)
            self.assertEqual(ports[0]["network_id"], None)
            self.assertEqual(ports[0]["tenant_id"], None)
            self.assertEqual(ports[0]["mac_address"], None)

    def test_port_diagnose_no_port_raises(self):
        with self._stubs(port=None):
            with self.assertRaises(n_exc.PortNotFound):
                self.plugin.diagnose_port(self.context.elevated(), 1, [])

    def test_port_diagnose_not_authorized(self):
        with self._stubs(port=None):
            with self.assertRaises(n_exc.NotAuthorized):
                self.plugin.diagnose_port(self.context, 1, [])


class TestPortDriverSelection(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, network=None, addr=None, mac=None,
               compat_map=None, driver_res=None, ipam="FOO"):
        network["ipam_strategy"] = "FOO"

        # Response from the backend driver
        self.expected_bridge = "backend-drivers-bridge"
        if driver_res is None:
            driver_res = {"uuid": 1, "bridge": self.expected_bridge}

        # Mock out the driver registry
        foo_driver = mock.Mock()
        foo_driver.create_port.return_value = driver_res
        foo_driver.select_ipam_strategy.return_value = "FOO"
        bar_driver = mock.Mock()
        bar_driver.create_port.return_value = driver_res
        bar_driver.select_ipam_strategy.return_value = "BAR"
        drivers = {"FOO": foo_driver,
                   "BAR": bar_driver}
        compat_map = compat_map or {}

        # Mock out the IPAM registry
        foo_ipam = mock.Mock()
        foo_ipam.allocate_ip_address.return_value = addr
        foo_ipam.allocate_mac_address.return_value = mac
        bar_ipam = mock.Mock()
        bar_ipam.allocate_ip_address.return_value = addr
        bar_ipam.allocate_mac_address.return_value = mac
        ipam = {"FOO": foo_ipam, "BAR": bar_ipam}

        with contextlib.nested(
            mock.patch("quark.db.api.port_create"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("oslo_utils.uuidutils.generate_uuid"),
            mock.patch("quark.plugin_views._make_port_dict"),
            mock.patch("quark.db.api.port_count_all"),
            mock.patch("neutron.quota.QuotaEngine.limit_check"),
            mock.patch("quark.plugin_modules.ports.registry."
                       "DRIVER_REGISTRY.drivers",
                       new_callable=mock.PropertyMock(return_value=drivers)),
            mock.patch("quark.plugin_modules.ports.registry."
                       "DRIVER_REGISTRY.port_driver_compat_map",
                       new_callable=mock.PropertyMock(
                           return_value=compat_map)),
            mock.patch("quark.plugin_modules.ports.ipam."
                       "IPAM_REGISTRY.strategies",
                       new_callable=mock.PropertyMock(return_value=ipam))
        ) as (port_create, net_find, gen_uuid, make_port,
              port_count, limit_check, _, _, _):
            net_find.return_value = network
            gen_uuid.return_value = 1
            port_count.return_value = 0
            yield (port_create, ipam, net_find)

    def test_create_port_with_bad_network_plugin_fails(self):
        network_dict = dict(id=1, tenant_id=self.context.tenant_id)
        port_name = "foobar"
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name))
        network = models.Network()
        network.update(network_dict)
        network["network_plugin"] = "FAIL"
        port_model = models.Port()
        port_model.update(port)
        port_models = port_model

        with self._stubs(network=network, addr=ip,
                         mac=mac) as (port_create, ipam, net_find):
            port_create.return_value = port_models

            exc = "Driver FAIL is not registered."
            with self.assertRaisesRegexp(n_exc.BadRequest, exc):
                self.plugin.create_port(self.context, port)

    def test_create_port_with_bad_port_network_plugin_fails(self):
        network_dict = dict(id=1, tenant_id=self.context.tenant_id)
        port_name = "foobar"
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, network_plugin="FAIL"))
        network = models.Network()
        network.update(network_dict)
        network["network_plugin"] = "FOO"
        port_model = models.Port()
        port_model.update(port)
        port_models = port_model

        with self._stubs(network=network, addr=ip,
                         mac=mac) as (port_create, ipam, net_find):
            port_create.return_value = port_models

            exc = "Driver FAIL is not registered."
            admin_ctx = self.context.elevated()
            with self.assertRaisesRegexp(n_exc.BadRequest, exc):
                self.plugin.create_port(admin_ctx, port)

    def test_create_port_with_incompatable_port_network_plugin_fails(self):
        network_dict = dict(id=1, tenant_id=self.context.tenant_id)
        port_name = "foobar"
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, network_plugin="BAR"))
        network = models.Network()
        network.update(network_dict)
        network["network_plugin"] = "FOO"
        port_model = models.Port()
        port_model.update(port)
        port_models = port_model

        with self._stubs(network=network, addr=ip,
                         mac=mac) as (port_create, ipam, net_find):
            port_create.return_value = port_models

            exc = ("Port driver BAR not allowed for underlying network "
                   "driver FOO.")
            admin_ctx = self.context.elevated()
            with self.assertRaisesRegexp(n_exc.BadRequest, exc):
                self.plugin.create_port(admin_ctx, port)

    def test_create_port_with_no_port_network_plugin(self):
        network = dict(id=1, tenant_id=self.context.tenant_id,
                       network_plugin="FOO")
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()

        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, device_owner="quark_tests",
                              bridge="quark_bridge", admin_state_up=False))

        expected_mac = "DE:AD:BE:EF:00:00"
        expected_bridge = "new_bridge"
        expected_device_owner = "new_device_owner"
        expected_admin_state = "new_state"

        port_create_dict = {}
        port_create_dict["port"] = port["port"].copy()
        port_create_dict["port"]["mac_address"] = expected_mac
        port_create_dict["port"]["device_owner"] = expected_device_owner
        port_create_dict["port"]["bridge"] = expected_bridge
        port_create_dict["port"]["admin_state_up"] = expected_admin_state

        admin_ctx = self.context.elevated()
        with self._stubs(network=network, addr=ip,
                         mac=mac) as (port_create, ipam, net_find):
            self.plugin.create_port(admin_ctx, port_create_dict)

            ipam["BAR"].allocate_mac_address.assert_not_called()
            ipam["BAR"].allocate_ip_address.assert_not_called()

            ipam["FOO"].allocate_ip_address.assert_called_once_with(
                admin_ctx, [], network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                segment_id=None, rollback=False, mac_address=mac)

            ipam["FOO"].allocate_mac_address.assert_called_once_with(
                admin_ctx, network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                mac_address=expected_mac, rollback=False,
                use_forbidden_mac_range=False)

            port_create.assert_called_once_with(
                admin_ctx, bridge=self.expected_bridge, uuid=1, name="foobar",
                admin_state_up=expected_admin_state, network_id=1,
                tenant_id="fake", id=1, device_owner=expected_device_owner,
                mac_address=mac["address"], device_id=2, backend_key=1,
                security_groups=[], addresses=[], instance_node_id='',
                network_plugin="FOO")

    def test_create_port_with_port_network_plugin(self):
        network = dict(id=1, tenant_id=self.context.tenant_id,
                       network_plugin="FOO")
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()

        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, device_owner="quark_tests",
                              bridge="quark_bridge", admin_state_up=False))

        expected_mac = "DE:AD:BE:EF:00:00"
        expected_bridge = "new_bridge"
        expected_device_owner = "new_device_owner"
        expected_admin_state = "new_state"
        expected_network_plugin = "FOO"

        port_create_dict = {}
        port_create_dict["port"] = port["port"].copy()
        port_create_dict["port"]["mac_address"] = expected_mac
        port_create_dict["port"]["device_owner"] = expected_device_owner
        port_create_dict["port"]["bridge"] = expected_bridge
        port_create_dict["port"]["admin_state_up"] = expected_admin_state
        port_create_dict["port"]["network_plugin"] = expected_network_plugin

        admin_ctx = self.context.elevated()
        with self._stubs(network=network, addr=ip,
                         mac=mac) as (port_create, ipam, net_find):
            self.plugin.create_port(admin_ctx, port_create_dict)

            ipam["BAR"].allocate_mac_address.assert_not_called()
            ipam["BAR"].allocate_ip_address.assert_not_called()

            ipam["FOO"].allocate_ip_address.assert_called_once_with(
                admin_ctx, [], network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                segment_id=None, rollback=False, mac_address=mac)

            ipam["FOO"].allocate_mac_address.assert_called_once_with(
                admin_ctx, network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                mac_address=expected_mac,
                rollback=False,
                use_forbidden_mac_range=False)

            port_create.assert_called_once_with(
                admin_ctx, bridge=self.expected_bridge, uuid=1, name="foobar",
                admin_state_up=expected_admin_state, network_id=1,
                tenant_id="fake", id=1, device_owner=expected_device_owner,
                mac_address=mac["address"], device_id=2, backend_key=1,
                security_groups=[], addresses=[], instance_node_id='',
                network_plugin=expected_network_plugin)

    def test_create_port_with_compatible_port_network_plugin(self):
        network = dict(id=1, tenant_id=self.context.tenant_id,
                       network_plugin="FOO")
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()

        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, device_owner="quark_tests",
                              bridge="quark_bridge", admin_state_up=False))

        expected_mac = "DE:AD:BE:EF:00:00"
        expected_bridge = "new_bridge"
        expected_device_owner = "new_device_owner"
        expected_admin_state = "new_state"
        expected_network_plugin = "BAR"

        port_create_dict = {}
        port_create_dict["port"] = port["port"].copy()
        port_create_dict["port"]["mac_address"] = expected_mac
        port_create_dict["port"]["device_owner"] = expected_device_owner
        port_create_dict["port"]["bridge"] = expected_bridge
        port_create_dict["port"]["admin_state_up"] = expected_admin_state
        port_create_dict["port"]["network_plugin"] = expected_network_plugin

        compat_map = {"BAR": ["FOO"]}
        admin_ctx = self.context.elevated()
        with self._stubs(network=network, addr=ip, mac=mac,
                         compat_map=compat_map) as (port_create, ipam,
                                                    net_find):
            self.plugin.create_port(admin_ctx, port_create_dict)

            ipam["FOO"].allocate_mac_address.assert_not_called()
            ipam["FOO"].allocate_ip_address.assert_not_called()

            ipam["BAR"].allocate_ip_address.assert_called_once_with(
                admin_ctx, [], network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                segment_id=None, rollback=False, mac_address=mac)

            ipam["BAR"].allocate_mac_address.assert_called_once_with(
                admin_ctx, network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                mac_address=expected_mac, rollback=False,
                use_forbidden_mac_range=False)

            port_create.assert_called_once_with(
                admin_ctx, bridge=self.expected_bridge, uuid=1, name="foobar",
                admin_state_up=expected_admin_state, network_id=1,
                tenant_id="fake", id=1, device_owner=expected_device_owner,
                mac_address=mac["address"], device_id=2, backend_key=1,
                security_groups=[], addresses=[], instance_node_id='',
                network_plugin=expected_network_plugin)

    def test_create_port_ipam_selection(self):
        network = dict(id=1, tenant_id=self.context.tenant_id,
                       network_plugin="FOO")
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()

        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, device_owner="quark_tests",
                              bridge="quark_bridge", admin_state_up=False))

        expected_mac = "DE:AD:BE:EF:00:00"
        expected_bridge = "new_bridge"
        expected_device_owner = "new_device_owner"
        expected_admin_state = "new_state"

        port_create_dict = {}
        port_create_dict["port"] = port["port"].copy()
        port_create_dict["port"]["mac_address"] = expected_mac
        port_create_dict["port"]["device_owner"] = expected_device_owner
        port_create_dict["port"]["bridge"] = expected_bridge
        port_create_dict["port"]["admin_state_up"] = expected_admin_state

        admin_ctx = self.context.elevated()
        with self._stubs(network=network, addr=ip,
                         mac=mac) as (port_create, ipam, net_find):
            self.plugin.create_port(admin_ctx, port_create_dict)

            ipam["BAR"].allocate_mac_address.assert_not_called()
            ipam["BAR"].allocate_ip_address.assert_not_called()

            ipam["FOO"].allocate_ip_address.assert_called_once_with(
                admin_ctx, [], network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                segment_id=None, rollback=False, mac_address=mac)

            ipam["FOO"].allocate_mac_address.assert_called_once_with(
                admin_ctx, network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                mac_address=expected_mac, rollback=False,
                use_forbidden_mac_range=False)

            port_create.assert_called_once_with(
                admin_ctx, bridge=self.expected_bridge, uuid=1, name="foobar",
                admin_state_up=expected_admin_state, network_id=1,
                tenant_id="fake", id=1, device_owner=expected_device_owner,
                mac_address=mac["address"], device_id=2, backend_key=1,
                security_groups=[], addresses=[], instance_node_id='',
                network_plugin="FOO")

    def test_create_port_ipam_selection_override_by_driver(self):
        network = dict(id=1, tenant_id=self.context.tenant_id,
                       network_plugin="BAR")
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()

        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, device_owner="quark_tests",
                              bridge="quark_bridge", admin_state_up=False))

        expected_mac = "DE:AD:BE:EF:00:00"
        expected_bridge = "new_bridge"
        expected_device_owner = "new_device_owner"
        expected_admin_state = "new_state"

        port_create_dict = {}
        port_create_dict["port"] = port["port"].copy()
        port_create_dict["port"]["mac_address"] = expected_mac
        port_create_dict["port"]["device_owner"] = expected_device_owner
        port_create_dict["port"]["bridge"] = expected_bridge
        port_create_dict["port"]["admin_state_up"] = expected_admin_state

        admin_ctx = self.context.elevated()
        with self._stubs(network=network, addr=ip,
                         mac=mac, ipam="BAR") as (port_create, ipam, net_find):
            self.plugin.create_port(admin_ctx, port_create_dict)

            ipam["FOO"].allocate_mac_address.assert_not_called()
            ipam["FOO"].allocate_ip_address.assert_not_called()

            ipam["BAR"].allocate_ip_address.assert_called_once_with(
                admin_ctx, [], network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                segment_id=None, rollback=False, mac_address=mac)

            ipam["BAR"].allocate_mac_address.assert_called_once_with(
                admin_ctx, network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                mac_address=expected_mac,
                rollback=False,
                use_forbidden_mac_range=False)

            port_create.assert_called_once_with(
                admin_ctx, bridge=self.expected_bridge, uuid=1, name="foobar",
                admin_state_up=expected_admin_state, network_id=1,
                tenant_id="fake", id=1, device_owner=expected_device_owner,
                mac_address=mac["address"], device_id=2, backend_key=1,
                security_groups=[], addresses=[], instance_node_id='',
                network_plugin="BAR")

    def test_create_port_network_plugin_response_no_uuid_raises(self):
        network_dict = dict(id=1, tenant_id=self.context.tenant_id)
        port_name = "foobar"
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name))
        network = models.Network()
        network.update(network_dict)
        network["network_plugin"] = "FOO"
        port_model = models.Port()
        port_model.update(port)
        port_models = port_model

        with self._stubs(network=network, addr=ip,
                         mac=mac, driver_res={}) as (port_create,
                                                     alloc_mac,
                                                     net_find):
            port_create.return_value = port_models

            exc = "uuid"
            with self.assertRaisesRegexp(KeyError, exc):
                self.plugin.create_port(self.context, port)

    def test_create_port_network_plugin_response_is_filtered(self):
        network = dict(id=1, tenant_id=self.context.tenant_id,
                       network_plugin="FOO")
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()

        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, device_owner="quark_tests",
                              bridge="quark_bridge", admin_state_up=False))

        expected_mac = "DE:AD:BE:EF:00:00"
        expected_bridge = "new_bridge"
        expected_device_owner = "new_device_owner"
        expected_admin_state = "new_state"

        port_create_dict = {}
        port_create_dict["port"] = port["port"].copy()
        port_create_dict["port"]["mac_address"] = expected_mac
        port_create_dict["port"]["device_owner"] = expected_device_owner
        port_create_dict["port"]["bridge"] = expected_bridge
        port_create_dict["port"]["admin_state_up"] = expected_admin_state

        driver_res = {
            "uuid": 5,
            "vlan_id": 50,
            "tags": [123, {"foo": "bar"}],
            "id": "fail",
            "randomkey": None
        }

        admin_ctx = self.context.elevated()
        with self._stubs(network=network, addr=ip,
                         mac=mac, driver_res=driver_res) as (port_create,
                                                             ipam,
                                                             net_find):
            self.plugin.create_port(admin_ctx, port_create_dict)

            ipam["BAR"].allocate_mac_address.assert_not_called()
            ipam["BAR"].allocate_ip_address.assert_not_called()

            ipam["FOO"].allocate_ip_address.assert_called_once_with(
                admin_ctx, [], network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                segment_id=None, rollback=False, mac_address=mac)

            ipam["FOO"].allocate_mac_address.assert_called_once_with(
                admin_ctx, network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                mac_address=expected_mac, rollback=False,
                use_forbidden_mac_range=False)

            port_create.assert_called_once_with(
                admin_ctx, bridge=expected_bridge, uuid=5, name="foobar",
                admin_state_up=expected_admin_state, network_id=1,
                tenant_id="fake", id=1, device_owner=expected_device_owner,
                mac_address=mac["address"], device_id=2, backend_key=5,
                security_groups=[], addresses=[], vlan_id=50,
                network_plugin=network["network_plugin"], instance_node_id='')


class TestQuarkPortCreateFiltering(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, network=None, addr=None, mac=None):
        network["network_plugin"] = "BASE"
        network["ipam_strategy"] = "ANY"

        with contextlib.nested(
            mock.patch("quark.db.api.port_create"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.ipam.QuarkIpam.allocate_ip_address"),
            mock.patch("quark.ipam.QuarkIpam.allocate_mac_address"),
            mock.patch("oslo_utils.uuidutils.generate_uuid"),
            mock.patch("quark.plugin_views._make_port_dict"),
            mock.patch("quark.db.api.port_count_all"),
            mock.patch("neutron.quota.QuotaEngine.limit_check")
        ) as (port_create, net_find, alloc_ip, alloc_mac, gen_uuid, make_port,
              port_count, limit_check):
            net_find.return_value = network
            alloc_ip.return_value = addr
            alloc_mac.return_value = mac
            gen_uuid.return_value = 1
            port_count.return_value = 0
            yield port_create, alloc_mac, net_find

    def test_create_port_attribute_filtering(self):
        network = dict(id=1, tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()
        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, device_owner="quark_tests",
                              bridge="quark_bridge", admin_state_up=False))

        port_create_dict = {}
        port_create_dict["port"] = port["port"].copy()
        port_create_dict["port"]["mac_address"] = "DE:AD:BE:EF:00:00"
        port_create_dict["port"]["device_owner"] = "ignored"
        port_create_dict["port"]["bridge"] = "ignored"
        port_create_dict["port"]["admin_state_up"] = "ignored"
        port_create_dict["port"]["network_plugin"] = "ignored"

        with self._stubs(network=network, addr=ip,
                         mac=mac) as (port_create, alloc_mac, net_find):
            self.plugin.create_port(self.context, port_create_dict)
            alloc_mac.assert_called_once_with(
                self.context, network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                mac_address=None, rollback=False,
                use_forbidden_mac_range=False)
            port_create.assert_called_once_with(
                self.context, addresses=[], network_id=network["id"],
                tenant_id="fake", uuid=1, name="foobar",
                mac_address=alloc_mac()["address"], backend_key=1, id=1,
                security_groups=[], network_plugin='BASE',
                device_id=2, instance_node_id='')

    def test_create_port_attribute_filtering_admin(self):
        network = dict(id=1, tenant_id=self.context.tenant_id)
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_name = "foobar"
        ip = dict()

        port = dict(port=dict(mac_address=mac["address"], network_id=1,
                              tenant_id=self.context.tenant_id, device_id=2,
                              name=port_name, device_owner="quark_tests",
                              bridge="quark_bridge", admin_state_up=False))

        expected_mac = "DE:AD:BE:EF:00:00"
        expected_bridge = "new_bridge"
        expected_device_owner = "new_device_owner"
        expected_admin_state = "new_state"
        expected_network_plugin = "BASE"

        port_create_dict = {}
        port_create_dict["port"] = port["port"].copy()
        port_create_dict["port"]["mac_address"] = expected_mac
        port_create_dict["port"]["device_owner"] = expected_device_owner
        port_create_dict["port"]["bridge"] = expected_bridge
        port_create_dict["port"]["admin_state_up"] = expected_admin_state
        port_create_dict["port"]["network_plugin"] = expected_network_plugin

        admin_ctx = self.context.elevated()
        with self._stubs(network=network, addr=ip,
                         mac=mac) as (port_create, alloc_mac, net_find):
            self.plugin.create_port(admin_ctx, port_create_dict)

            alloc_mac.assert_called_once_with(
                admin_ctx, network["id"], 1,
                cfg.CONF.QUARK.ipam_reuse_after,
                mac_address=expected_mac,
                rollback=False,
                use_forbidden_mac_range=False)

            port_create.assert_called_once_with(
                admin_ctx, bridge=expected_bridge, uuid=1, name="foobar",
                admin_state_up=expected_admin_state, network_id=1,
                tenant_id="fake", id=1, device_owner=expected_device_owner,
                mac_address=mac["address"], device_id=2, backend_key=1,
                security_groups=[], addresses=[],
                network_plugin=expected_network_plugin, instance_node_id='')


class TestQuarkPortUpdateFiltering(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self):
        with contextlib.nested(
            mock.patch("quark.db.api.port_find"),
            mock.patch("quark.db.api.port_update"),
            mock.patch("quark.drivers.registry.DriverRegistry.get_driver"),
            mock.patch("quark.plugin_views._make_port_dict"),
            mock.patch("neutron.quota.QuotaEngine.limit_check")
        ) as (port_find, port_update, get_driver, make_port, limit_check):
            yield port_find, port_update

    def test_update_port_attribute_filtering(self):
        new_port = {}
        new_port["port"] = {
            "mac_address": "DD:EE:FF:00:00:00", "device_owner": "new_owner",
            "bridge": "new_bridge", "admin_state_up": False, "device_id": 3,
            "network_id": 10, "backend_key": 1234, "name": "new_name",
            "network_plugin": "BASE"}

        with self._stubs() as (port_find, port_update):
            self.plugin.update_port(self.context, 1, new_port)
            port_update.assert_called_once_with(
                self.context,
                port_find(),
                name="new_name",
                security_groups=[])

    def test_update_port_attribute_filtering_admin(self):
        new_port = {}
        new_port["port"] = {
            "mac_address": "DD:EE:FF:00:00:00", "device_owner": "new_owner",
            "bridge": "new_bridge", "admin_state_up": False, "device_id": 3,
            "network_id": 10, "backend_key": 1234, "name": "new_name",
            "network_plugin": "BASE"}

        admin_ctx = self.context.elevated()
        with self._stubs() as (port_find, port_update):
            self.plugin.update_port(admin_ctx, 1, new_port)
            port_update.assert_called_once_with(
                admin_ctx,
                port_find(),
                name="new_name",
                bridge=new_port["port"]["bridge"],
                admin_state_up=new_port["port"]["admin_state_up"],
                device_owner=new_port["port"]["device_owner"],
                mac_address=new_port["port"]["mac_address"],
                device_id=new_port["port"]["device_id"],
                security_groups=[])


class TestQuarkPortCreateAsAdvancedService(test_quark_plugin.TestQuarkPlugin):

    @contextlib.contextmanager
    def _stubs(self, port=None, network=None, addr=None, mac=None):
        if network:
            network["network_plugin"] = "BASE"
            network["ipam_strategy"] = "ANY"
        port_model = models.Port()
        port_model.update(port['port'])
        port_models = port_model
        db_mod = "quark.db.api"
        ipam = "quark.ipam.QuarkIpam"
        with contextlib.nested(
            mock.patch("%s.port_create" % db_mod),
            mock.patch("%s.network_find" % db_mod),
            mock.patch("%s.port_find" % db_mod),
            mock.patch("%s.allocate_ip_address" % ipam),
            mock.patch("%s.allocate_mac_address" % ipam),
            mock.patch("%s.port_count_all" % db_mod),
        ) as (port_create, net_find, port_find, alloc_ip, alloc_mac,
              port_count):
            port_create.return_value = port_models
            net_find.return_value = network
            port_find.return_value = None
            alloc_ip.return_value = addr
            alloc_mac.return_value = mac
            port_count.return_value = 0
            yield port_create

    def test_advanced_service_create_port_other_tenant_network(self):
        """NCP-1819 - Advanced service can create port on any network

        Tests when an advanced service creating a port on another tenant's
        network does not fail AND the tenant_id is that of the context's.
        """
        self.context.is_advsvc = True
        network_id = "foobar"
        network = dict(id=network_id,
                       tenant_id="other_tenant")
        ip = dict()
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_1 = dict(port=dict(mac_address="AA:BB:CC:DD:EE:00",
                                network_id=network_id,
                                tenant_id=self.context.tenant_id, device_id=5,
                                name="Fake"))

        with self._stubs(port=port_1, network=network, addr=ip, mac=mac):
            port = self.plugin.create_port(self.context, port_1)
            self.assertEqual(self.context.tenant_id, port['tenant_id'])

    def test_advsvc_can_create_port_with_another_tenant_id(self):
        """NCP-1819 - Advanced Service can create port on another tenant's net

        Tests that an advanced service can create a port on another tenant's
        network.
        """
        another_tenant_id = 'im-another-tenant'
        self.context.is_advsvc = True
        network_id = "foobar"
        network = dict(id=network_id,
                       tenant_id="other_tenant")
        ip = dict()
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_1 = dict(port=dict(mac_address="AA:BB:CC:DD:EE:00",
                                network_id=network_id,
                                tenant_id=another_tenant_id, device_id=5,
                                name="Fake"))

        with self._stubs(port=port_1, network=network, addr=ip, mac=mac):
            port = self.plugin.create_port(self.context, port_1)
            self.assertEqual(another_tenant_id, port['tenant_id'])

    def test_non_advsvc_cannot_create_port_another_network(self):
        """NCP-1819 - Normal tenant port create should fail another's network

        Tests that a normal tenant creating a port on another tenant's network
        should not be allowed and throws an exception.
        """
        normal_tenant_id = "other_tenant"
        network_id = "foobar"
        network = dict(id=network_id,
                       tenant_id=normal_tenant_id)
        ip = dict()
        mac = dict(address="AA:BB:CC:DD:EE:FF")
        port_1 = dict(port=dict(mac_address="AA:BB:CC:DD:EE:00",
                                network_id=network_id,
                                tenant_id=normal_tenant_id, device_id=5,
                                name="Fake"))

        with self._stubs(port=port_1, network=network, addr=ip, mac=mac):
            with self.assertRaises(n_exc.NotAuthorized):
                self.plugin.create_port(self.context, port_1)



# Copyright 2016 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from neutron_lib import exceptions as n_exc


class NetworkAlreadyExists(n_exc.Conflict):
    message = _("Network %(id)s already exists.")


class InvalidMacAddressRange(n_exc.NeutronException):
    message = _("Invalid MAC address range %(cidr)s.")


class InvalidEthertype(n_exc.NeutronException):
    message = _("Invalid Ethertype %(ethertype)s.")


class MacAddressRangeNotFound(n_exc.NotFound):
    message = _("MAC address range %(mac_address_range_id)s not found.")


class MacAddressRangeInUse(n_exc.InUse):
    message = _("MAC address range %(mac_address_range_id)s in use.")


class MacAddressNotFound(n_exc.NotFound):
    message = _("MAC address %(mac_address_id)s (%(readable_mac)s) not found.")


class InvalidSegmentAllocationRange(n_exc.NeutronException):
    message = _("Invalid segment allocation range: %(msg)s.")


class SegmentAllocationRangeNotFound(n_exc.NotFound):
    message = _(
        "Segment allocation range %(segment_allocation_range_id)s not found.")


class SegmentAllocationRangeInUse(n_exc.InUse):
    message = _(
        "Segment allocation range %(segment_allocation_range_id)s in use.")


class SegmentAllocationFailure(n_exc.Conflict):
    message = _("No more segment ids available for segment "
                "type:%(segment_type)s id:%(segment_id)s.")


class RouteNotFound(n_exc.NotFound):
    message = _("Route %(route_id)s not found.")


class AmbiguousNetworkId(n_exc.InvalidInput):
    msg = _("Segment ID required for network %(net_id)s.")


class AmbigiousLswitchCount(n_exc.NeutronException):
    message = _("Too many lswitches for network %(net_id)s.")


class IpAddressNotFound(n_exc.NotFound):
    message = _("IP Address %(addr_id)s not found.")


class PortRequiresDisassociation(n_exc.BadRequest):
    message = _("Port requires disassociation before IP can be deleted")


class RouteConflict(n_exc.NeutronException):
    message = _("Route overlaps existing route %(route_id)s with %(cidr)s")


class DuplicateRouteConflict(n_exc.NeutronException):
    message = _("More than one default route found for subnet %(subnet_id)s")


class InvalidPhysicalNetworkType(n_exc.NeutronException):
    message = _("Providernet type %(net_type)s is invalid")


class SegmentIdUnsupported(n_exc.NeutronException):
    message = _("Segmentation ID is unsupported for network type %(net_type)s")


class SegmentIdRequired(n_exc.NeutronException):
    message = _("Segmentation ID is required for network type %(net_type)s")


class PhysicalNetworkNotFound(n_exc.NeutronException):
    message = _("Physical network %(phys_net)s not found!")


class InvalidIpamStrategy(n_exc.BadRequest):
    message = _("IPAM Strategy %(strat)s is invalid.")


class ProvidernetParamError(n_exc.NeutronException):
    message = _("%(msg)s")


class BadNVPState(n_exc.NeutronException):
    message = _("No networking information found for network %(net_id)s")


class IPAddressRetryableFailure(n_exc.NeutronException):
    message = _("Allocation of %(ip_addr)s for net %(net_id)s failed, "
                "retrying...")


class IPAddressPolicyRetryableFailure(IPAddressRetryableFailure):
    message = _("Allocation of %(ip_addr)s for net %(net_id)s failed "
                "due to policy retrying...")


class IPAddressNotInSubnet(n_exc.InvalidInput):
    message = _("Requested IP %(ip_addr)s not in subnet %(subnet_id)s")


class IPAddressProhibitedByPolicy(n_exc.InvalidInput):
    message = _("IP %(ip_addr)s is prohibited by policies on the subnet")


class ProviderNetworkOutOfIps(n_exc.NeutronException):
    message = _("Network %(net_id)s appears to be out of IP Addresses."
                "You may want to try again in a few seconds")


class IPPolicyNotFound(n_exc.NeutronException):
    message = _("IP Policy %(id)s not found.")


class IPPolicyAlreadyExists(n_exc.NeutronException):
    message = _("IP Policy %(id)s already exists for %(n_id)s")


class IPPolicyInUse(n_exc.InUse):
    message = _("IP allocation policy %(id)s in use.")


class DriverLimitReached(n_exc.InvalidInput):
    message = _("Driver has reached limit on resource '%(limit)s'")


class SecurityGroupsNotImplemented(n_exc.InvalidInput):
    message = _("Security Groups are not currently implemented on port "
                "create")


class TenantNetworkSecurityGroupRulesNotEnabled(n_exc.InvalidInput):
    message = _("Tenant network security group rules are not currently "
                "allowed by environment_capabilities configuration.")


class EgressSecurityGroupRulesNotEnabled(n_exc.InvalidInput):
    message = _("Egress security group rules are not currently allowed "
                "by environment_capabilities configuration.")


class SecurityGroupsRequireDevice(n_exc.InvalidInput):
    message = _("Security Groups may only be applied to ports connected to "
                "devices")


class RedisConnectionFailure(n_exc.NeutronException):
    message = _("No connection to Redis could be made.")


class NoBackendConnectionsDefined(n_exc.NeutronException):
    message = _("This driver cannot be used without a backend connection "
                "definition. %(msg)s")


class FloatingIpNotFound(n_exc.NotFound):
    message = _("Floating IP %(id)s not found.")


class ScalingIpNotFound(n_exc.NotFound):
    message = _("Scaling IP %(id)s not found.")


class RemoveFloatingIpFailure(n_exc.NeutronException):
    message = _("An error occurred when trying to remove the "
                "floating IP %(id)s.")


class RegisterFloatingIpFailure(n_exc.NeutronException):
    message = _("An error occurred when trying to register the floating IP "
                "%(id)s.")


class PortAlreadyContainsFloatingIp(n_exc.Conflict):
    message = _("Port %(port_id)s already has an associated floating IP.")


class FixedIpAllocatedToGatewayIp(n_exc.Conflict):
    message = _("Gateway IP of Floating IP subnet conflicts with Fixed IP "
                "allocated to port %(port_id)s of network %(network_id)s.")


class FixedIpDoesNotExistsForPort(n_exc.BadRequest):
    message = _("Fixed IP %(fixed_ip)s does not exist on Port %(port_id)s")


class PortAlreadyContainsScalingIp(n_exc.Conflict):
    message = _("Port %(port_id)s already has an associated scaling IP.")


class NoAvailableFixedIpsForPort(n_exc.Conflict):
    message = _("There are no available fixed IPs for port %(port_id)s")


class PortDoesNotHaveAGateway(n_exc.Conflict):
    message = _("Port %(port_id)s does not have a gateway")


class PortAlreadyAssociatedToFloatingIp(n_exc.BadRequest):
    message = _("Port %(port_id)s is already associated with "
                "floating IP %(flip_id)s")


class FloatingIpUpdateNoPortIdSupplied(n_exc.BadRequest):
    message = _("When no port is currently associated to the floating IP, "
                "port_id is required but was not supplied")


class PortOrDeviceNotFound(n_exc.PortNotFound):
    message = _("Suitable port or device could not be found")


class NotAllPortOrDeviceFound(n_exc.NotFound):
    message = _("Not all ports or devices from request could be found")


class CannotAddMoreIPsToPort(n_exc.OverQuota):
    message = _("Cannot add more IPs to port")


class CannotCreateMoreSharedIPs(n_exc.OverQuota):
    message = _("Cannot create more shared IPs on selected network")


class JobNotFound(n_exc.NotFound):
    message = _("Job %(job_id)s not found")



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
v2 Neutron Plug-in API Quark Implementation
"""

from neutron.extensions import securitygroup as sg_ext
from neutron import neutron_plugin_base_v2
from neutron.quota import resource as qres
from neutron.quota import resource_registry as qres_reg
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging

from quark.api import extensions
from quark import ip_availability
from quark.plugin_modules import floating_ips
from quark.plugin_modules import ip_addresses
from quark.plugin_modules import ip_policies
from quark.plugin_modules import jobs
from quark.plugin_modules import mac_address_ranges
from quark.plugin_modules import networks
from quark.plugin_modules import ports
from quark.plugin_modules import router
from quark.plugin_modules import routes
from quark.plugin_modules import security_groups
from quark.plugin_modules import segment_allocation_ranges
from quark.plugin_modules import subnets

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

quark_resources = [
    qres.BaseResource('alloc_pools_per_subnet',
                      'quota_alloc_pools_per_subnet'),
    qres.BaseResource('dns_nameservers_per_subnet',
                      'quota_dns_nameservers_per_subnet'),
    qres.BaseResource('ports_per_network',
                      'quota_ports_per_network'),
    qres.BaseResource('routes_per_subnet',
                      'quota_routes_per_subnet'),
    qres.BaseResource('security_rules_per_group',
                      'quota_security_rules_per_group'),
    qres.BaseResource('security_groups_per_port',
                      'quota_security_groups_per_port'),
    qres.BaseResource('v4_subnets_per_network',
                      'quota_v4_subnets_per_network'),
    qres.BaseResource('v6_subnets_per_network',
                      'quota_v6_subnets_per_network'),
    qres.BaseResource('fixed_ips_per_port',
                      'quota_fixed_ips_per_port')
]

quark_quota_opts = [
    cfg.IntOpt("quota_alloc_pools_per_subnet",
               default=5,
               help=_("Maximum number of allocation pools per subnet")),
    cfg.IntOpt('quota_dns_nameservers_per_subnet',
               default=2,
               help=_('Maximum number of dns nameservers per subnet')),
    cfg.IntOpt('quota_ports_per_network',
               default=250,
               help=_('Maximum ports per network')),
    cfg.IntOpt('quota_routes_per_subnet',
               default=3,
               help=_('Maximum routes per subnet')),
    cfg.IntOpt('quota_security_rules_per_group',
               default=20,
               help=_('Maximum security group rules in a group')),
    cfg.IntOpt("quota_security_groups_per_port",
               default=5,
               help=_("Maximum number of security groups per port")),
    cfg.IntOpt('quota_v4_subnets_per_network',
               default=1,
               help=_('Maximum v4 subnets per network')),
    cfg.IntOpt('quota_v6_subnets_per_network',
               default=1,
               help=_('Maximum v6 subnets per network')),
    cfg.IntOpt('quota_fixed_ips_per_port',
               default=6,
               help=_('Maximum number of fixed IPs per port')),
]


def append_quark_extensions(conf):
    """Adds the Quark API Extensions to the extension path.

    Pulled out for test coveage.
    """
    if 'api_extensions_path' in conf:
        conf.set_override('api_extensions_path', ":".join(extensions.__path__))

append_quark_extensions(CONF)


CONF.register_opts(quark_quota_opts, "QUOTAS")
qres_reg.ResourceRegistry.get_instance().register_resources(quark_resources)


def sessioned(func):
    def _wrapped(self, context, *args, **kwargs):
        res = func(self, context, *args, **kwargs)
        if not context.session.is_active:
            context.session.close()
            # NOTE(mdietz): Forces neutron to get a fresh session
            #              if it needs it after our call
            context._session = None
        return res
    return _wrapped


class Plugin(neutron_plugin_base_v2.NeutronPluginBaseV2,
             sg_ext.SecurityGroupPluginBase):
    supported_extension_aliases = ["mac_address_ranges", "routes",
                                   "ip_addresses", "security-group",
                                   "diagnostics", "subnets_quark",
                                   "provider", "ip_policies", "quotas",
                                   "networks_quark", "router",
                                   "ip_availabilities", "ports_quark",
                                   "floatingip", "segment_allocation_ranges",
                                   "scalingip", "jobs"]

    def __init__(self):
        LOG.info("Starting quark plugin")

    def _fix_missing_tenant_id(self, context, body, key):
        """Will add the tenant_id to the context from body.

        It is assumed that the body must have a tenant_id because neutron
        core could never have gotten here otherwise.
        """
        if not body:
            raise n_exc.BadRequest(resource=key,
                                   msg="Body malformed")
        resource = body.get(key)
        if not resource:
            raise n_exc.BadRequest(resource=key,
                                   msg="Body malformed")
        if context.tenant_id is None:
            context.tenant_id = resource.get("tenant_id")
        if context.tenant_id is None:
            msg = _("Running without keystone AuthN requires "
                    "that tenant_id is specified")
            raise n_exc.BadRequest(resource=key, msg=msg)

    @sessioned
    def get_mac_address_range(self, context, id, fields=None):
        return mac_address_ranges.get_mac_address_range(context, id, fields)

    @sessioned
    def get_mac_address_ranges(self, context):
        return mac_address_ranges.get_mac_address_ranges(context)

    @sessioned
    def create_mac_address_range(self, context, mac_range):
        self._fix_missing_tenant_id(context, mac_range, "mac_address_range")
        return mac_address_ranges.create_mac_address_range(context, mac_range)

    @sessioned
    def delete_mac_address_range(self, context, id):
        mac_address_ranges.delete_mac_address_range(context, id)

    @sessioned
    def create_security_group(self, context, security_group):
        self._fix_missing_tenant_id(context, security_group, "security_group")
        return security_groups.create_security_group(context, security_group)

    @sessioned
    def create_security_group_rule(self, context, security_group_rule):
        self._fix_missing_tenant_id(
            context, security_group_rule, "security_group_rule")
        return security_groups.create_security_group_rule(context,
                                                          security_group_rule)

    @sessioned
    def delete_security_group(self, context, id):
        security_groups.delete_security_group(context, id)

    @sessioned
    def delete_security_group_rule(self, context, id):
        security_groups.delete_security_group_rule(context, id)

    @sessioned
    def get_security_group(self, context, id, fields=None):
        return security_groups.get_security_group(context, id, fields)

    @sessioned
    def get_security_group_rule(self, context, id, fields=None):
        return security_groups.get_security_group_rule(context, id, fields)

    @sessioned
    def get_security_groups(self, context, filters=None, fields=None,
                            sorts=None, limit=None, marker=None,
                            page_reverse=False):
        return security_groups.get_security_groups(context, filters, fields,
                                                   sorts, limit, marker,
                                                   page_reverse)

    @sessioned
    def get_security_group_rules(self, context, filters=None, fields=None,
                                 sorts=None, limit=None, marker=None,
                                 page_reverse=False):
        return security_groups.get_security_group_rules(context, filters,
                                                        fields, sorts, limit,
                                                        marker, page_reverse)

    @sessioned
    def update_security_group(self, context, id, security_group):
        return security_groups.update_security_group(context, id,
                                                     security_group)

    @sessioned
    def create_ip_policy(self, context, ip_policy):
        self._fix_missing_tenant_id(context, ip_policy, "ip_policy")
        return ip_policies.create_ip_policy(context, ip_policy)

    @sessioned
    def get_ip_policy(self, context, id):
        return ip_policies.get_ip_policy(context, id)

    @sessioned
    def get_ip_policies(self, context, **filters):
        return ip_policies.get_ip_policies(context, **filters)

    @sessioned
    def update_ip_policy(self, context, id, ip_policy):
        return ip_policies.update_ip_policy(context, id, ip_policy)

    @sessioned
    def delete_ip_policy(self, context, id):
        return ip_policies.delete_ip_policy(context, id)

    @sessioned
    def get_ip_addresses(self, context, **filters):
        return ip_addresses.get_ip_addresses(context, **filters)

    @sessioned
    def get_ip_address(self, context, id):
        return ip_addresses.get_ip_address(context, id)

    @sessioned
    def create_ip_address(self, context, ip_address):
        self._fix_missing_tenant_id(context, ip_address, "ip_address")
        return ip_addresses.create_ip_address(context, ip_address)

    @sessioned
    def update_ip_address(self, context, id, ip_address):
        return ip_addresses.update_ip_address(context, id, ip_address)

    @sessioned
    def delete_ip_address(self, context, id):
        return ip_addresses.delete_ip_address(context, id)

    @sessioned
    def create_port(self, context, port):
        self._fix_missing_tenant_id(context, port, "port")
        return ports.create_port(context, port)

    @sessioned
    def get_port(self, context, id, fields=None):
        return ports.get_port(context, id, fields)

    @sessioned
    def update_port(self, context, id, port):
        return ports.update_port(context, id, port)

    @sessioned
    def update_port_for_ip(self, context, ip_id, id, port):
        return ip_addresses.update_port_for_ip_address(context, ip_id, id,
                                                       port)

    @sessioned
    def get_ports(self, context, limit=None, page_reverse=False, sorts=None,
                  marker=None, filters=None, fields=None):
        return ports.get_ports(context, limit, sorts, marker, page_reverse,
                               filters, fields)

    @sessioned
    def get_ports_for_ip_address(self, context, ip, limit=None,
                                 page_reverse=False, sorts=None, marker=None,
                                 filters=None, fields=None):
        return ip_addresses.get_ports_for_ip_address(context, ip, limit, sorts,
                                                     marker, page_reverse,
                                                     filters, fields)

    @sessioned
    def get_port_for_ip_address(self, context, ip_id, id, fields=None):
        return ip_addresses.get_port_for_ip_address(context, ip_id, id, fields)

    @sessioned
    def get_ports_count(self, context, filters=None):
        return ports.get_ports_count(context, filters)

    @sessioned
    def delete_port(self, context, id):
        return ports.delete_port(context, id)

    @sessioned
    def disassociate_port(self, context, id, ip_address_id):
        return ports.disassociate_port(context, id, ip_address_id)

    @sessioned
    def diagnose_port(self, context, id, fields):
        return ports.diagnose_port(context, id, fields)

    @sessioned
    def get_route(self, context, id):
        return routes.get_route(context, id)

    @sessioned
    def get_routes(self, context):
        return routes.get_routes(context)

    @sessioned
    def create_route(self, context, route):
        self._fix_missing_tenant_id(context, route, "route")
        return routes.create_route(context, route)

    @sessioned
    def delete_route(self, context, id):
        routes.delete_route(context, id)

    @sessioned
    def create_subnet(self, context, subnet):
        self._fix_missing_tenant_id(context, subnet, "subnet")
        return subnets.create_subnet(context, subnet)

    @sessioned
    def update_subnet(self, context, id, subnet):
        return subnets.update_subnet(context, id, subnet)

    @sessioned
    def get_subnet(self, context, id, fields=None):
        return subnets.get_subnet(context, id, fields)

    @sessioned
    def get_subnets(self, context, limit=None, page_reverse=False, sorts=None,
                    marker=None, filters=None, fields=None):
        return subnets.get_subnets(context, limit, page_reverse, sorts, marker,
                                   filters, fields)

    @sessioned
    def get_subnets_count(self, context, filters=None):
        return subnets.get_subnets_count(context, filters)

    @sessioned
    def delete_subnet(self, context, id):
        return subnets.delete_subnet(context, id)

    @sessioned
    def diagnose_subnet(self, context, id, fields):
        return subnets.diagnose_subnet(context, id, fields)

    @sessioned
    def create_network(self, context, network):
        self._fix_missing_tenant_id(context, network, "network")
        return networks.create_network(context, network)

    @sessioned
    def update_network(self, context, id, network):
        return networks.update_network(context, id, network)

    @sessioned
    def get_network(self, context, id, fields=None):
        return networks.get_network(context, id, fields)

    @sessioned
    def get_networks(self, context, limit=None, sorts=None, marker=None,
                     page_reverse=False, filters=None, fields=None):
        return networks.get_networks(context, limit, sorts, marker,
                                     page_reverse, filters, fields)

    @sessioned
    def get_networks_count(self, context, filters=None):
        return networks.get_networks_count(context, filters)

    @sessioned
    def delete_network(self, context, id):
        return networks.delete_network(context, id)

    @sessioned
    def diagnose_network(self, context, id, fields):
        return networks.diagnose_network(context, id, fields)

    # NOTE(mdietz): we don't actually support these, but despite the fact that
    #               they're extensions in Neutron, Nova still expects to be
    #               able to call some of these as if they aren't
    def create_router(self, context, router):
        raise NotImplementedError()

    def update_router(self, context, id, router):
        raise NotImplementedError()

    def get_router(self, context, id, fields=None):
        return router.get_router(context, id, fields)

    def delete_router(self, context, id):
        raise NotImplementedError()

    def get_routers(self, context, filters=None, fields=None,
                    sorts=None, limit=None, marker=None, page_reverse=False):
        return router.get_routers(context, filters=filters, fields=fields,
                                  sorts=sorts, limit=limit, marker=marker,
                                  page_reverse=page_reverse)

    def add_router_interface(self, context, router_id, interface_info):
        raise NotImplementedError()

    def remove_router_interface(self, context, router_id, interface_info):
        raise NotImplementedError()

    @sessioned
    def create_floatingip(self, context, floatingip):
        self._fix_missing_tenant_id(context, floatingip, "floatingip")
        return floating_ips.create_floatingip(context,
                                              floatingip["floatingip"])

    @sessioned
    def update_floatingip(self, context, id, floatingip):
        return floating_ips.update_floatingip(context, id,
                                              floatingip["floatingip"])

    @sessioned
    def get_floatingip(self, context, id, fields=None):
        return floating_ips.get_floatingip(context, id, fields)

    @sessioned
    def delete_floatingip(self, context, id):
        return floating_ips.delete_floatingip(context, id)

    @sessioned
    def get_floatingips(self, context, filters=None, fields=None,
                        sorts=None, limit=None, marker=None,
                        page_reverse=False):
        return floating_ips.get_floatingips(context, filters=filters,
                                            fields=fields, sorts=sorts,
                                            limit=limit, marker=marker,
                                            page_reverse=page_reverse)

    def get_routers_count(self, context, filters=None):
        raise NotImplementedError()

    def get_floatingips_count(self, context, filters=None):
        return floating_ips.get_floatingips_count(context, filters)

    def get_ip_availability(self, **kwargs):
        return ip_availability.get_ip_availability(**kwargs)

    @sessioned
    def get_segment_allocation_range(self, context, id, fields=None):
        return segment_allocation_ranges.get_segment_allocation_range(
            context, id, fields)

    @sessioned
    def get_segment_allocation_ranges(self, context, **filters):
        return segment_allocation_ranges.get_segment_allocation_ranges(
            context, **filters)

    @sessioned
    def create_segment_allocation_range(self, context, sa_range):
        self._fix_missing_tenant_id(
            context, sa_range, "segment_allocation_range")
        return segment_allocation_ranges.create_segment_allocation_range(
            context, sa_range)

    @sessioned
    def delete_segment_allocation_range(self, context, id):
        segment_allocation_ranges.delete_segment_allocation_range(
            context, id)

    def create_scalingip(self, context, scalingip):
        self._fix_missing_tenant_id(context, scalingip, "scalingip")
        return floating_ips.create_scalingip(context, scalingip["scalingip"])

    @sessioned
    def update_scalingip(self, context, id, scalingip):
        return floating_ips.update_scalingip(context, id,
                                             scalingip["scalingip"])

    @sessioned
    def get_scalingip(self, context, id, fields=None):
        return floating_ips.get_scalingip(context, id, fields)

    @sessioned
    def delete_scalingip(self, context, id):
        return floating_ips.delete_scalingip(context, id)

    @sessioned
    def get_scalingips(self, context, filters=None, fields=None,
                       sorts=None, limit=None, marker=None,
                       page_reverse=False):
        return floating_ips.get_scalingips(context, filters=filters,
                                           fields=fields, sorts=sorts,
                                           limit=limit, marker=marker,
                                           page_reverse=page_reverse)

    @sessioned
    def get_jobs(self, context, **filters):
        return jobs.get_jobs(context, **filters)

    @sessioned
    def get_job(self, context, id):
        return jobs.get_job(context, id)

    @sessioned
    def create_job(self, context, job):
        self._fix_missing_tenant_id(context, job, 'job')
        return jobs.create_job(context, job)

    @sessioned
    def update_job(self, context, id, job):
        self._fix_missing_tenant_id(context, job, 'job')
        return jobs.update_job(context, id, job)

    @sessioned
    def delete_job(self, context, id):
        return jobs.delete_job(context, id)



# Copyright 2015 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc


class TagValidationError(Exception):
    def __init__(self, value, message):
        self.value = value
        self.message = message


class Tag(object):

    @classmethod
    def get_name(cls):
        """API name of the tag."""
        if not hasattr(cls, 'NAME'):
            raise NotImplementedError()
        return cls.NAME

    @classmethod
    def get_prefix(cls):
        """Tag 'key', saved in the database as <prefix>:<value>"""
        return "%s:" % cls.get_name().upper()

    def serialize(self, value):
        return "%s%s" % (self.get_prefix(), value)

    def deserialize(self, tag):
        if self.is_tag(tag):
            try:
                return tag[len(self.get_prefix()):]
            except Exception:
                pass
        return None

    def validate(self, value):
        raise NotImplementedError

    def set(self, model, value):
        """Set tag on model object."""
        self.validate(value)
        self._pop(model)
        value = self.serialize(value)
        model.tags.append(value)

    def get(self, model):
        """Get a matching valid tag off the model."""
        for tag in model.tags:
            if self.is_tag(tag):
                value = self.deserialize(tag)
                try:
                    self.validate(value)
                    return value
                except TagValidationError:
                    continue
        return None

    def _pop(self, model):
        """Pop all matching tags off the model and return them."""
        tags = []

        # collect any exsiting tags with matching prefix
        for tag in model.tags:
            if self.is_tag(tag):
                tags.append(tag)

        # remove collected tags from model
        if tags:
            for tag in tags:
                model.tags.remove(tag)

        return tags

    def pop(self, model):
        """Pop all matching tags off the port, return a valid one."""
        tags = self._pop(model)
        if tags:
            for tag in tags:
                value = self.deserialize(tag)
                try:
                    self.validate(value)
                    return value
                except TagValidationError:
                    continue

    def is_tag(self, tag):
        """Is a given tag this type?"""
        return tag[0:len(self.get_prefix())] == self.get_prefix()

    def has_tag(self, model):
        """Does the given port have this tag?"""
        for tag in model.tags:
            if self.is_tag(tag):
                return True
        return False


class VlanTag(Tag):

    NAME = "vlan_id"
    MIN_VLAN_ID = 1
    MAX_VLAN_ID = 4096

    def validate(self, value):
        """Validates a VLAN ID.

        :param value: The VLAN ID to validate against.
        :raises TagValidationError: Raised if the VLAN ID is invalid.
        """
        try:
            vlan_id_int = int(value)
            assert vlan_id_int >= self.MIN_VLAN_ID
            assert vlan_id_int <= self.MAX_VLAN_ID
        except Exception:
            msg = ("Invalid vlan_id. Got '%(vlan_id)s'. "
                   "vlan_id should be an integer between %(min)d and %(max)d "
                   "inclusive." % {'vlan_id': value,
                                   'min': self.MIN_VLAN_ID,
                                   'max': self.MAX_VLAN_ID})
            raise TagValidationError(value, msg)
        return True


class TagRegistry(object):

    tags = {}

    def get_all(self, model):
        """Get all known tags from a model.

        Returns a dict of {<tag_name>:<tag_value>}.
        """
        tags = {}
        for name, tag in self.tags.items():
            for mtag in model.tags:
                if tag.is_tag(mtag):
                    tags[name] = tag.get(model)
        return tags

    def set_all(self, model, **tags):
        """Validate and set all known tags on a port."""
        for name, tag in self.tags.items():
            if name in tags:
                value = tags.pop(name)
                if value:
                    try:
                        tag.set(model, value)
                    except TagValidationError as e:
                        raise n_exc.BadRequest(
                            resource="tags",
                            msg="%s" % (e.message))


class PortTagRegistry(TagRegistry):

    def __init__(self):
        self.tags = {
            VlanTag.get_name(): VlanTag()
        }


PORT_TAG_REGISTRY = PortTagRegistry()



# Copyright (c) 2013 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from neutron.api import extensions
from neutron import manager
from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
LOG = logging.getLogger(__name__)


class Diagnostician(object):
    def __init__(self, plugin):
        self.plugin = plugin

    def diag_not_implemented(self, res, id, input):
        LOG.warning("Diagnostics not implemented on resource %ss." % res)
        raise n_exc.ServiceUnavailable()

    def diagnose(self, res, input, req, id):
        LOG.debug("Requested diagnostics fields %s on resource %s with id %s"
                  % (input['diag'], res, id))
        return getattr(
            self.plugin, 'diagnose_%s' % res.replace('-', '_'),
            functools.partial(self.diag_not_implemented, res))(
                req.context, id, input['diag'])


class Diagnostics(extensions.ExtensionDescriptor):
    def get_name(self):
        return "Diagnostics"

    def get_alias(self):
        return "diagnostics"

    def get_description(self):
        return "Diagnostics extension"

    def get_namespace(self):
        return "None"

    def get_updated(self):
        return "never"

    def get_actions(self):
        diagnose = Diagnostician(manager.NeutronManager.get_plugin()).diagnose
        resources = ['port', 'subnet', 'network']
        return (extensions.ActionExtension('%ss' % res, 'diag',
                functools.partial(diagnose, res)) for res in resources)
